from .commons import wrap_val
from .structures import Structure, _get_all_fields_by_name
from .fields import StructureClass, Map, String, OneOf, Boolean
from .serialization import FunctionCall, deserialize_structure, serialize


class Deserializer(Structure):
    """
    A high level API for a deserializer: from a dict or anything else that could be sent as a JSON, to
    a :class:`Structure`.
    The advantage of this over the lower level function is that it is more explicit and self-validating.
    In other words, it prevents the user from creating an invalid mapper.

    Arguments:
        target_class(:class:`StructureClass`):
            A class extending the abstract :class:`Structure` that this deserializer is build for.
            Example:

            .. code-block:: python

                class Foo(Structure):
                    id = Integer
                    name = String

               Deserializer(target_class=Foo)


        mapper(dict): optional
            The key is the target attribute name. The value can either be a path of the value in the source dict
            using dot notation, for example: "aaa.bbb", or a :class:`FunctionCall`. In the latter case,
            the function is the used to preprocess the input prior to deserialization/validation.
            The args attribute in the function call is optional. If non provided, the input to the function is
            the value with the same key. Otherwise it is the keys of the values in the input that are injected
            to the provided function. See working examples in the tests link above.
            This class will ensure that the mapper is a valid one for its target_class.
            Example:

            .. code-block:: python

                 class Foo(Structure):
                    m = Map
                    s = String
                    i = Integer

                mapper = {
                    "m": "a.b",
                    "s": FunctionCall(func=lambda x: f'the string is {x}', args=['name.first']),
                    'i': FunctionCall(func=operator.add, args=['i', 'j'])
                }
                Deserializer(target_class=Foo, mapper = mapper).deserializer(the_input_dict)

        camel_case_convert(bool): Optional
            If true, will convert any camelCase key that does not have explicit mapping to a snake_case attribute
            name. Default is False.

    """

    target_class = StructureClass
    mapper = Map[String, OneOf[String, FunctionCall, Map]]
    camel_case_convert = Boolean(default=False)

    _required = ["target_class"]

    def __validate__(self):
        valid_keys = set(_get_all_fields_by_name(self.target_class).keys())
        if self.mapper:
            for key in self.mapper:
                if key.split(".")[0] not in valid_keys:
                    raise ValueError(
                        f"Invalid key in mapper for class {self.target_class.__name__}: {key}. Keys must be one of "
                        "the class fields. "
                    )

    def deserialize(self, input_data, keep_undefined=None):
        return deserialize_structure(
            self.target_class,
            input_data,
            mapper=self.mapper,
            keep_undefined=keep_undefined,
            camel_case_convert=self.camel_case_convert,
        )


class Serializer(Structure):
    """
           A high level API for a serializer: from an instance of :class:`Structure`, to something
           that can be sent as a Json(usually a dict). The advantage of this over the lower level
           function is that it is more explicit andself-validating.
           In other words, it prevents the user from creating an invalid mapper.

           Arguments:
               source(:class:`Structure`):
                   An instance of :class:`Structure` that this serializer is build for.
                   Example:

                   .. code-block:: python

                       class Foo(Structure):
                           i = Integer
                           f = Float

                       foo = Foo(f=5.5, i=999)
                       Serializer(source=foo)

               mapper(dict): optional
                   The key is the target key name. The value can either be a path of the value in the source object
                   using dot notation, for example: "aaa.bbb", or a :class:`FunctionCall`. In the latter case,
                   the function is the used to preprocess the input prior to deserialization/validation. \
                   The args attribute in the function call is optional. If non provided, the input to the function is
                   the attribute with the same key. Otherwise it is the names of the attributes  in the input
                   that are injected to the provided function. \

                   See working examples in the tests link above. \
                   This class will ensure that the mapper is a valid one for its target_class.
                   Example:

                   .. code-block:: python

                        class Foo(Structure):
                            f = Float
                            i = Integer

                        foo = Foo(f=5.5, i=999)
                        mapper = {
                            'output_floats': FunctionCall(func=lambda f: [int(f)], args=['i']),
                            'output_int': FunctionCall(func=lambda x: str(x), args=['f'])
                        }
                        assert Serializer(source=foo, mapper=mapper).serialize() == {
                             'output_floats': [999],
                             'output_int': '5.5'
                        }


    """

    source = Structure
    mapper = Map[String, OneOf[String, FunctionCall, Map]]

    _required = ["source"]

    def __validate__(self):
        def verify_key_in_mapper(key, valid_keys, source_class):
            if key.split(".")[0] not in valid_keys:
                raise ValueError(
                    f"Invalid key in mapper for class {source_class.__name__}: {key}. Keys must be one of the class "
                    "fields. "
                )
            if isinstance(self.mapper[key], (FunctionCall,)):
                args = self.mapper[key].args
                if isinstance(args, (list,)):
                    for arg in args:
                        if arg not in valid_keys:
                            raise ValueError(
                                f"Mapper[{key}] has a function call with an invalid argument: {arg}"
                            )

        source_class = self.source.__class__
        valid_keys = set(_get_all_fields_by_name(source_class).keys())
        if self.mapper:
            for key in self.mapper:
                verify_key_in_mapper(key, valid_keys, source_class)

    def serialize(self, compact: bool = True, camel_case_convert: bool = False):
        """

        Arguments:
              compact(boolean): optional
                    whether to use a compact form for Structure that is a simple wrapper of a field.
                    for example: if a Structure has only one field of an int, if compact is True
                    it will serialize the structure as an int instead of a dictionary.
                    Default is False.

              camel_case_convert(dict): optional
                    If True, convert any camel-case key that does not have a mapping in the mapper to a snake-case
                    attribute.
                    Default is False.
        """
        return serialize(
            self.source,
            mapper=self.mapper,
            compact=compact,
            camel_case_convert=camel_case_convert,
        )


def deserializer_by_discriminator(class_by_discriminator_value, keep_undefined=False):
    """
    create deserialized based on discriminator value in the input.

    :param class_by_discriminator_value: a dictionary of the Structure class by the discriminator value
    :return: A function that should be used in a mapper, for which the first argument is the key for
             The discriminator field, and the second is the key for the data field that is deserialized based
             on the provided class_by_discriminator_value.
    """
    _desererializer_by_type = {
        k: Deserializer(t) for (k, t) in class_by_discriminator_value.items()
    }

    def _get_content(discriminator, data):
        if discriminator not in _desererializer_by_type:
            raise ValueError(
                f"discriminator: got {wrap_val(discriminator)}; Expected one of {list(_desererializer_by_type.keys())}"
            )

        return _desererializer_by_type[discriminator].deserialize(
            data, keep_undefined=keep_undefined
        )

    return _get_content



import json
from typing import Union

from .commons import wrap_val
from .structures import Structure

MISSING_VALUES = "missing values"
ADDITIONAL_VALUES = "additional values"
DIFFERENT_ORDER = "different location"


def _add_val(container: dict, key: str, diff):
    if key in container:
        container[key].append(diff)
    else:
        container[key] = [diff]


def _diff_set(val, otherval) -> dict:
    result = {}
    for v in val:
        if v not in otherval:
            _add_val(result, MISSING_VALUES, v)
    for v in otherval:
        if v not in val:
            _add_val(result, ADDITIONAL_VALUES, v)
    return result


def _diff_dict(val, otherval) -> dict:
    result = {}
    for key, v in val.items():
        if key not in otherval:
            _add_val(result, MISSING_VALUES, key)
        elif isinstance(v, Structure):
            diff = _find_diff(v, otherval[key])
            if diff:
                result[key] = diff
        elif isinstance(otherval[key], Structure):
            diff = _find_diff(otherval[key], v)
            if diff:
                result[key] = diff
        else:
            result[key] = f"{v} vs {otherval[key]}"
    for key, v in otherval.items():
        if key not in val:
            _add_val(result, ADDITIONAL_VALUES, key)
        elif isinstance(v, Structure):
            diff = _find_diff(val[key], v)
            if diff:
                result[key] = diff
        else:
            result[key] = f"{v} vs {otherval[key]}"
    return result


def _diff_list(
    val, otherval, outer_result: dict, outer_key: str
) -> dict:  # pylint: disable=too-many-branches
    result = {}
    for i, v in enumerate(val):
        if v == otherval[i]:
            continue
        try:
            index = otherval.index(v)
            msg = f"index {i} vs {index}"
            _add_val(result, DIFFERENT_ORDER, msg)
        except ValueError:
            diff = _find_diff(v, otherval[i])
            if diff:
                if outer_key:
                    outer_result[f"{outer_key}[{i}]"] = diff
                else:
                    result[i] = diff
            else:
                internal_diff = _find_diff(
                    v, otherval[i], outer_result=outer_result, out_key=outer_key
                )
                if internal_diff:
                    if outer_key:
                        outer_result[f"{outer_key}[{i}]"] = internal_diff
                    else:
                        result[i] = internal_diff
    for i, v in enumerate(otherval):
        if v == val[i]:
            continue
        try:
            val.index(v)
            continue
        except ValueError:
            if outer_key and f"{outer_key}[{i}]" in outer_result or i in result:
                continue
            diff = _find_diff(v, val[i])
            if diff:
                if outer_key:
                    outer_result[f"{outer_key}[{i}]"] = diff
                else:
                    result[i] = diff
            else:
                internal_diff = _find_diff(
                    v, otherval[i], outer_result=outer_result, out_key=outer_key
                )
                if internal_diff:
                    if outer_key:
                        outer_result[f"{outer_key}[{i}]"] = internal_diff
                    else:
                        result[i] = internal_diff
    return result


def find_diff(first, second) -> Union[dict, str]:
    """
    Utility for testing to find the differences between two values that are "supposed"
    to be equal. This is useful to have more useful assertion error messages,
    especially with pytest, using pytest_assertrepr_compare.

    Arguments:
        first:
            first value. Can be a Structure, list, dict, set
        second:
            second value. Can be a Structure, list, dict, set

    Returns:
        a dictionary with the differences, or the difference is trivial - a string.
        Note that this function does not employ any sophisticated algorithm and is meant
        just as a best-effort utility for testing.

    Example of the output (taken from the unit tests):

    .. code-block:: python

        actual = find_diff(bar2, bar1)
        assert actual == {
            "f": {
                "m['aaa']": {"age": "123 vs 12"},
                "missing values": ["bbb"],
                "additional values": ["ccc"],
            },
            "missing values": ["x"],
        }

    """
    return _find_diff(first, second)


def _find_diff(struct, other, outer_result=None, out_key=None) -> Union[dict, str]:

    if struct.__class__ != other.__class__:
        return {"class": f"{struct.__class__} vs. {other.__class__}"}
    if isinstance(struct, (list, tuple)):
        res_val = _diff_list(
            struct, other, outer_result=outer_result, outer_key=out_key
        )
        if res_val and out_key:
            outer_result[out_key] = res_val
        return res_val
    elif isinstance(struct, dict):
        res_val = _diff_dict(struct, other)
        if res_val and outer_result:
            for i in res_val:
                if i not in {MISSING_VALUES, ADDITIONAL_VALUES}:
                    outer_result[f"{out_key}[{wrap_val(i)}]"] = res_val[i]
                else:
                    outer_result[i] = res_val[i]
        return res_val
    elif isinstance(struct, set):
        res_val = _diff_set(struct, other)
        if res_val and out_key:
            outer_result[out_key] = res_val
        return res_val

    internal_props = ["_instantiated"]
    res = {}
    if isinstance(struct, Structure):  #  pylint: disable=too-many-nested-blocks
        for k, val in sorted(struct.__dict__.items()):
            if k not in internal_props:
                if k not in other.__dict__:
                    _add_val(res, MISSING_VALUES, k)
                elif val != other.__dict__.get(k):
                    otherval = other.__dict__.get(k)
                    if isinstance(val, Structure):
                        res[k] = _find_diff(val, otherval)
                    elif isinstance(val, (list, tuple, set, dict)):
                        if val.__class__ != otherval.__class__:
                            res[k] = {
                                "class": f"{val.__class__} vs. {otherval.__class__}"
                            }
                        elif len(val) != len(otherval):
                            res[k] = f"length of {len(val)} vs {len(otherval)}"
                        else:
                            if isinstance(val, (list, tuple)):
                                res_val = _diff_list(
                                    val, otherval, outer_result=res, outer_key=k
                                )
                                if res_val:
                                    res[k] = res_val
                            elif isinstance(val, dict):
                                res_val = _diff_dict(val, otherval)
                                if res_val:
                                    for i in res_val:
                                        if i not in {MISSING_VALUES, ADDITIONAL_VALUES}:
                                            res[f"{k}[{wrap_val(i)}]"] = res_val[i]
                                        else:
                                            # _add_val(res, ADDITIONAL_VALUES, k)
                                            res[i] = res_val[i]
                            elif isinstance(val, set):
                                res_val = _diff_set(val, otherval)
                                if res_val:
                                    res[k] = res_val

                    else:
                        res[k] = _find_diff(val, otherval)
        for k, val in sorted(other.__dict__.items()):
            if k not in internal_props:
                if k not in struct.__dict__:
                    _add_val(res, ADDITIONAL_VALUES, k)
    else:
        if struct != other:
            return f"{struct} vs {other}"

    return res


def pytest_assertrepr_compare(op, left, right):
    if isinstance(left, Structure) and isinstance(right, Structure) and op == "==":
        res = [
            "found the following differences between the structures:",
            json.dumps(_find_diff(left, right)),
        ]
        return res



import collections.abc
import enum
import importlib.util
import inspect
import logging
import os
import sys
import builtins
import typing
from os.path import relpath
from pathlib import Path
import ast

from .commons import wrap_val, nested
from .fields import AllOf, AnyOf, FunctionCall, Map, OneOf
from .enum import Enum
from .serialization_wrappers import Deserializer, Serializer
from .structures import (
    ImmutableStructure,
    NoneField,
    Structure,
    Field,
    get_typing_lib_info,
)
from .types_ast import get_imports, get_models, models_to_src
from .utility import type_is_generic

builtins_types = [
    getattr(builtins, k)
    for k in dir(builtins)
    if isinstance(getattr(builtins, k), type)
]

module = getattr(inspect, "__class__")

INDENT = " " * 4

AUTOGEN_NOTE = [
    "",
    "#### This stub was autogenerated by Typedpy",
    "###########################################",
    "",
]


_private_to_public_pkg = {"werkzeug.localxxx": "flask"}


def _get_package(v, attrs):
    pkg_name = attrs.get("__package__") or "%^$%^$%^#"
    if v.startswith(pkg_name):
        return v[len(pkg_name) :]
    return _private_to_public_pkg.get(v, v)


def _as_something(k, attrs):
    return f" as {k}" if attrs.get("__file__", "").endswith("__init__.py") else ""


def _get_anyof_typing(field, locals_attrs, additional_classes):
    union_fields = getattr(field, "_fields", [])
    if len(union_fields) == 2 and isinstance(union_fields[1], NoneField):
        info = _get_type_info(union_fields[0], locals_attrs, additional_classes)
        return f"Optional[{info}] = None"

    fields = ",".join(
        [_get_type_info(f, locals_attrs, additional_classes) for f in union_fields]
    )
    return f"Union[{fields}]"


def _get_type_info_for_typing_generic(
    the_type, locals_attrs, additional_classes
) -> typing.Optional[str]:
    origin = getattr(the_type, "__origin__", None)
    args = getattr(the_type, "__args__", [])
    if origin in {list, set, tuple}:
        if len(args) != 1:
            return origin.__name__
        return f"{origin.__name__}[{_get_type_info(args[0], locals_attrs, additional_classes)}]"
    mapped_args = (
        [_get_type_info(a, locals_attrs, additional_classes) for a in args]
        if args
        else []
    )

    if origin is collections.abc.Callable:
        args_st = (
            ""
            if not mapped_args
            else f"[{mapped_args[0]}]"
            if len(mapped_args) == 1
            else f"[[{','.join(mapped_args[:-1])}], {mapped_args[-1]}]"
        )
        additional_classes.add(typing.Callable)
        return f"Callable{args_st}"
    if origin is dict:
        if mapped_args:
            return f"dict[{', '.join(mapped_args)}]"
        return "dict"
    if origin is collections.abc.Iterable:
        additional_classes.add(typing.Iterable)
        args_st = "" if not mapped_args else f"[{mapped_args[0]}]"
        return f"Iterable{args_st}"

    if origin is collections.abc.Mapping:
        additional_classes.add(typing.Mapping)
        args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
        return f"Mapping{args_st}"

    if origin is type:
        args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
        return f"Type{args_st}"

    if origin is collections.abc.Iterator:
        additional_classes.add(typing.Iterator)
        args_st = "" if not mapped_args else f"[{mapped_args[0]}]"
        return f"Iterator{args_st}"

    if origin is typing.Union:
        if "None" in mapped_args:
            mapped_args = [a for a in mapped_args if str(a) != "None"]
            the_type = "Optional"
        else:
            the_type = "Type"
        args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
        return f"{the_type}{args_st}"

    return None


def _found_in_local_attrs(field, attrs):
    for k, v in attrs.items():
        if v is field:
            return k
    return None


def _get_type_info(field, locals_attrs, additional_classes):
    try:
        if field is ...:
            return "..."
        found_import_key = _found_in_local_attrs(field, locals_attrs)
        if found_import_key:
            return found_import_key
        if isinstance(field, (list, dict)):
            cls_name = field.__class__.__name__
            mapped_args = [
                _get_type_info(a, locals_attrs, additional_classes) for a in field
            ]
            args_st = "" if not mapped_args else f"[{', '.join(mapped_args)}]"
            return f"{cls_name}{args_st}"

        if field in [None, type(None)]:
            return "None"
        if isinstance(field, (AnyOf, OneOf, AllOf)):
            return _get_anyof_typing(field, locals_attrs, additional_classes)

        if isinstance(field, Map):
            if not field.items:
                return "dict"
            sub_types = ", ".join(
                [
                    _get_type_info(f, locals_attrs, additional_classes)
                    for f in field.items
                ]
            )
            return f"dict[{sub_types}]"

        if isinstance(field, Enum):
            if field._is_enum:
                additional_classes.add(field._enum_class)
                return field._enum_class.__name__

        the_type = (
            getattr(field, "get_type", field) if isinstance(field, Field) else field
        )
        # deal with from __future__ import annotations
        if isinstance(the_type, str):
            the_type = eval(the_type, locals_attrs)
            field = the_type
            if isinstance(the_type, str):
                return the_type

        if the_type in builtins_types:
            return the_type.__name__

        if the_type in [typing.Any, typing.Optional, typing.Union, typing.NoReturn]:
            return getattr(the_type, "_name")
        if (
            getattr(the_type, "__module__", None)
            and the_type.__module__ != "builtins"
            and the_type not in locals_attrs
            and not type_is_generic(the_type)
        ):
            additional_classes.add(field)

        if type_is_generic(the_type):
            res = _get_type_info_for_typing_generic(
                the_type, locals_attrs, additional_classes
            )
            if res:
                return res

            return _get_type_info(
                get_typing_lib_info(field), locals_attrs, additional_classes
            )

        return the_type.__name__
    except Exception as e:
        logging.exception(e)
        return "Any"


def _get_all_type_info(cls, locals_attrs, additional_classes) -> dict:
    type_by_name = {}
    required = getattr(cls, "_required", None)
    for field_name, field in cls.get_all_fields_by_name().items():
        try:
            type_info_str: str = _get_type_info(field, locals_attrs, additional_classes)
            if (
                field_name not in required
                and required is not None
                and not type_info_str.startswith("Optional[")
            ):
                type_info_str = f"Optional[{type_info_str}] = None"
            type_by_name[field_name] = type_info_str
        except Exception as e:
            logging.exception(e)

    return type_by_name


def _get_struct_classes(attrs, only_calling_module=True):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isclass(v)
            and issubclass(v, Structure)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
            and v not in {Deserializer, Serializer, ImmutableStructure, FunctionCall}
        )
    }


def _get_imported_classes(attrs):
    def _valid_module(v):
        return hasattr(v, "__module__") and attrs["__name__"] != v.__module__

    res = []
    for k, v in attrs.items():
        if (
            not k.startswith("_")
            and (isinstance(v, module) or _valid_module(v))
            and not _is_internal_sqlalchemy(v)
        ):
            if isinstance(v, module):
                if v.__name__ != k:
                    parts = v.__name__.split(".")
                    if len(parts) > 1:
                        first_parts = parts[:-1]
                        last_part = parts[-1]
                        res.append(
                            (
                                k,
                                f"from {'.'.join(first_parts)} import {last_part} as {k}",
                            )
                        )
                    elif nested(lambda: getattr(v.os, k)) == v:
                        res.append((k, f"from os import {k} as {k}"))
                else:
                    res.append((k, f"import {k}"))
            else:
                pkg = _get_package(v.__module__, attrs)

                name_in_pkg = getattr(v, "__name__", k)
                name_to_import = name_in_pkg if name_in_pkg and name_in_pkg != k else k
                as_import = (
                    _as_something(k, attrs) if name_to_import == k else f" as {k}"
                )
                import_stmt = f"from {pkg} import {name_to_import}{as_import}"
                if pkg == "__future__":
                    res = [(k, import_stmt)] + res
                else:
                    res.append((k, import_stmt))
    return res


def _get_ordered_args(unordered_args: dict):
    optional_args = {k: v for k, v in unordered_args.items() if v.endswith("= None")}
    mandatory_args = {k: v for k, v in unordered_args.items() if k not in optional_args}
    return {**mandatory_args, **optional_args}


def _get_mapped_extra_imports(additional_imports) -> dict:
    mapped = {}
    for c in additional_imports:
        try:
            name = _get_type_info(c, {}, set())
            if inspect.isclass(c) and issubclass(c, Structure):
                module_name = c.__module__
            else:
                module_name = (
                    c.get_type.__module__
                    if isinstance(c, Field)
                    else c.__module__
                    if name != "Any"
                    else None
                )
            if module_name:
                mapped[name] = module_name
        except Exception as e:
            logging.exception(e)
    return mapped


def _is_sqlalchemy(attr):
    module_name = getattr(attr, "__module__", "")
    return module_name and (
        module_name.startswith("sqlalchemy.orm")
        or module_name.startswith("sqlalchemy.sql")
    )


def _is_internal_sqlalchemy(attr):
    module_name = getattr(attr, "__module__", "")
    return module_name in {"sqlalchemy.orm.decl_api", "sqlalchemy.sql.schema"}


def _try_extract_column_type(attr):
    if attr.__class__.__name__ == "InstrumentedAttribute":
        return next(iter(attr.expression.base_columns)).type.python_type


def _skip_sqlalchemy_attribute(attribute):
    return attribute.startswith("_") or attribute in {"registry", "metadata"}


def _get_method_and_attr_list(cls, members):
    all_fields = cls.get_all_fields_by_name() if issubclass(cls, Structure) else {}
    ignored_methods = (
        dir(Structure)
        if issubclass(cls, Structure)
        else dir(Field)
        if issubclass(cls, Structure)
        else dir(enum)
        if issubclass(cls, enum.Enum)
        else {}
    )
    private_prefix = "_" if issubclass(cls, enum.Enum) else "__"
    method_list = []
    attrs = []
    cls_dict = cls.__dict__
    for attribute in members:
        if attribute.startswith(private_prefix):
            continue
        attr = (
            cls_dict.get(attribute)
            if attribute in cls_dict
            else getattr(cls, attribute, None)
        )
        if _is_sqlalchemy(attr):
            if not _skip_sqlalchemy_attribute(attribute):
                members[attribute] = _try_extract_column_type(attr)
                attrs.append(attribute)
            continue
        is_func = not inspect.isclass(attr) and (
            callable(attr) or isinstance(attr, (property, classmethod, staticmethod))
        )
        if not any(
            [
                attr is None,
                inspect.isclass(attr),
                is_func,
                isinstance(attr, cls),
                issubclass(cls, Structure),
            ]
        ):
            attrs.append(attribute)
            continue

        if is_func and attribute not in all_fields and attribute not in ignored_methods:
            method_list.append(attribute)

    if (
        not issubclass(cls, Structure)
        and not issubclass(cls, enum.Enum)
        and "__init__" in members
    ):
        method_list = ["__init__"] + method_list

    for name in cls_dict.get("__annotations__", {}):
        if name not in attrs and not issubclass(cls, Structure):
            attrs.append(name)
    return method_list, attrs


def _is_sqlalchemy_orm_model(cls):
    return nested(lambda: str(cls.__class__.__module__), "").startswith(
        "sqlalchemy.orm"
    )


def _get_sqlalchemy_init(attributes_with_type):
    res = [f"def __init__(self, *,"]
    for p, p_type in attributes_with_type:
        res.append(f"{INDENT}{p}: {p_type} = None,")
    res.append(f"{INDENT}**kw")
    res.append(f"): ...")
    return res


def _get_methods_info(cls, locals_attrs, additional_classes) -> list:
    method_by_name = []
    members = {}
    members.update(dict(cls.__dict__))
    annotations = cls.__dict__.get("__annotations__", {})
    for a in annotations:
        members[a] = annotations[a]
    members.update(annotations)

    method_list, cls_attrs = _get_method_and_attr_list(cls, members)
    attributes_with_type = []
    for attr in cls_attrs:
        the_type = members.get(attr, None)
        if getattr(the_type, "__module__", "").startswith("_") or nested(
            lambda: the_type.__class__.__name__, ""
        ).startswith("_"):
            # private class/module
            continue
        if inspect.isclass(the_type) or type_is_generic(the_type):
            resolved_type = (
                _get_type_info(the_type, locals_attrs, additional_classes)
                if the_type
                else "Any"
            )
        else:
            resolved_type = (
                _get_type_info(the_type.__class__, locals_attrs, additional_classes)
                if the_type is not None
                else "Any"
            )
        attributes_with_type.append((attr, resolved_type))
    method_by_name = [
        f"{attr}: {resolved_type}" for (attr, resolved_type) in attributes_with_type
    ]
    cls_dict = cls.__dict__

    for name in method_list:
        method_cls = members[name].__class__
        is_property = False
        func = cls_dict.get(name) if name in cls_dict else getattr(cls, name, None)
        func = (
            getattr(cls, name)
            if isinstance(func, (classmethod, staticmethod, property))
            and not _is_sqlalchemy(func)
            else func
        )
        if isinstance(func, property):
            is_property = True
            func = func.__get__
        try:
            sig = inspect.signature(func)
            return_annotations = (
                ""
                if sig.return_annotation == inspect._empty
                else f" -> {_get_type_info(sig.return_annotation, locals_attrs, additional_classes)}"
            )
            params_by_name = []
            if _is_sqlalchemy_orm_model(cls) and name == "__init__":
                method_by_name.extend(_get_sqlalchemy_init(attributes_with_type))
                continue

            if method_cls is classmethod:
                params_by_name.append(("cls", ""))
            if is_property:
                params_by_name.append(("self", ""))
            found_last_positional = False
            arg_position = 0
            for p, v in sig.parameters.items():
                if is_property and arg_position < 2:
                    continue
                arg_position += 1
                optional_globe = (
                    "**"
                    if v.kind == inspect.Parameter.VAR_KEYWORD
                    else "*"
                    if v.kind == inspect.Parameter.VAR_POSITIONAL
                    else ""
                )
                if v.kind == inspect.Parameter.VAR_POSITIONAL:
                    found_last_positional = True
                if (
                    v.kind == inspect.Parameter.KEYWORD_ONLY
                    and not found_last_positional
                ):
                    params_by_name.append(("*", ""))
                    found_last_positional = True
                default = (
                    ""
                    if v.default == inspect._empty
                    else f" = {v.default.__name__}"
                    if inspect.isclass(v.default)
                    else " = None"
                )
                type_annotation = (
                    ""
                    if v.annotation == inspect._empty
                    else f": {_get_type_info(v.annotation, locals_attrs, additional_classes)}"
                )
                p_name = f"{optional_globe}{p}"
                type_annotation = (
                    type_annotation[: -len(" = None")]
                    if (type_annotation.endswith("= None") and default)
                    else type_annotation
                )
                params_by_name.append((p_name, f"{type_annotation}{default}"))

            params_as_str = ", ".join([f"{k}{v}" for (k, v) in params_by_name])
            method_by_name.append("")
            if method_cls is staticmethod:
                method_by_name.append("@staticmethod")
            elif method_cls is classmethod:
                method_by_name.append("@classmethod")
            elif is_property:
                method_by_name.append("@property")
            method_by_name.append(
                f"def {name}({params_as_str}){return_annotations}: ..."
            )
        except Exception as e:
            logging.warning(e)
            method_by_name.append(f"def {name}(self, *args, **kw): ...")

    return method_by_name


def _get_init(cls, ordered_args: dict) -> str:
    init_params = f",\n{INDENT * 2}".join(
        [f"{INDENT * 2}self"] + [f"{k}: {v}" for k, v in ordered_args.items()]
    )
    kw_opt = (
        f",\n{INDENT * 2}**kw" if getattr(cls, "_additionalProperties", True) else ""
    )
    return f"    def __init__(\n{init_params}{kw_opt}\n{INDENT}): ..."


def _get_shallow_clone(cls, ordered_args: dict) -> str:
    ordered_args_with_none = {}
    for k, v in ordered_args.items():
        ordered_args_with_none[k] = v if v.endswith("= None") else f"{v} = None"
    params = f",\n{INDENT * 2}".join(
        [f"{INDENT * 2}self"] + [f"{k}: {v}" for k, v in ordered_args_with_none.items()]
    )
    kw_opt = (
        f",\n{INDENT * 2}**kw" if getattr(cls, "_additionalProperties", True) else ""
    )
    return f"    def shallow_clone_with_overrides(\n{params}{kw_opt}\n{INDENT}): ..."


def get_stubs_of_structures(
    struct_classe_by_name: dict, local_attrs, additional_classes
) -> list:
    out_src = []
    for cls_name, cls in struct_classe_by_name.items():
        fields_info = _get_all_type_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )
        method_info = _get_methods_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )

        ordered_args = _get_ordered_args(fields_info)
        out_src.append(f"class {cls_name}(Structure):")
        if not fields_info and not method_info:
            out_src.append(f"{INDENT}pass")
            out_src.append("")
            continue

        out_src.append(_get_init(cls, ordered_args))
        out_src.append("")
        out_src.append(_get_shallow_clone(cls, ordered_args))
        out_src.append("")

        for field_name, type_name in ordered_args.items():
            out_src.append(f"    {field_name}: {type_name}")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def get_stubs_of_enums(
    enum_classes_by_name: dict, local_attrs, additional_classes
) -> list:
    out_src = []
    for cls_name, cls in enum_classes_by_name.items():

        method_info = _get_methods_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )

        out_src.append(f"class {cls_name}(enum.Enum):")
        for v in cls:
            out_src.append(f"{INDENT}{v.name} = enum.auto()")
        out_src.append("")

        out_src.append("")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def add_imports(local_attrs: dict, additional_classes, existing_imports: set) -> list:
    base_typing = ["Union", "Optional", "Any", "TypeVar", "Type", "NoReturn"]
    typing_types_to_import = [t for t in base_typing if t not in existing_imports]
    base_import_statements = []
    if typing_types_to_import:
        base_import_statements.append(
            f"from typing import {', '.join(typing_types_to_import)}"
        )
    base_import_statements += [
        "from typedpy import Structure",
        "",
    ]
    extra_imports_by_name = _get_mapped_extra_imports(additional_classes)
    extra_imports = {
        f"from {_get_package(v, local_attrs)} import {k}{_as_something(k, local_attrs)}"
        for k, v in extra_imports_by_name.items()
        if (
            (
                k not in local_attrs
                or local_attrs[k].__module__ != local_attrs["__name__"]
            )
            and k not in existing_imports
        )
    }
    return base_import_statements + sorted(extra_imports)


def _get_enum_classes(attrs, only_calling_module):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isclass(v)
            and issubclass(v, enum.Enum)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
        )
    }


def _get_other_classes(attrs, only_calling_module):
    res = {}
    for k, v in attrs.items():
        if (
            inspect.isclass(v)
            and not issubclass(v, enum.Enum)
            and not issubclass(v, Structure)
        ):
            res[k] = v

    return res


def _get_functions(attrs, only_calling_module):
    return {
        k: v
        for k, v in attrs.items()
        if (
            inspect.isfunction(v)
            and (v.__module__ == attrs["__name__"] or not only_calling_module)
        )
    }


def _get_type_annotation(
    prefix: str, annotation, default: str, local_attrs, additional_classes
):
    def _correct_for_return_annotation(res: str):
        if "->" in prefix:
            return res[:-7] if res.endswith("= None") else res
        if default:
            cleaned_res = res[: -len(" = None")] if res.endswith("= None") else res
            return f"{cleaned_res}{default}"
        return res

    try:
        res = (
            ""
            if annotation == inspect._empty
            else f"{prefix}{_get_type_info(annotation, local_attrs, additional_classes)}"
        )
        return _correct_for_return_annotation(res)
    except Exception as e:
        logging.exception(e)
        return ""


def get_stubs_of_functions(func_by_name, local_attrs, additional_classes) -> list:
    def _convert_default(d):
        return d.__name__ if inspect.isclass(d) else None

    out_src = []
    for name, func in func_by_name.items():
        sig = inspect.signature(func)
        return_annotations = _get_type_annotation(
            " -> ", sig.return_annotation, "", local_attrs, additional_classes
        )
        params_by_name = []
        found_last_positional = False
        for p, v in sig.parameters.items():
            default = (
                ""
                if v.default == inspect._empty
                else f" = {_convert_default(v.default)}"
            )
            type_annotation = _get_type_annotation(
                ": ", v.annotation, default, local_attrs, additional_classes
            )
            optional_globe = (
                "**"
                if v.kind == inspect.Parameter.VAR_KEYWORD
                else "*"
                if v.kind == inspect.Parameter.VAR_POSITIONAL
                else ""
            )
            if v.kind == inspect.Parameter.VAR_POSITIONAL:
                found_last_positional = True
            if v.kind == inspect.Parameter.KEYWORD_ONLY and not found_last_positional:
                params_by_name.append(("*", ""))
                found_last_positional = True
            p_name = f"{optional_globe}{p}"
            params_by_name.append((p_name, type_annotation))
        params_as_str = ", ".join([f"{k}{v}" for (k, v) in params_by_name])

        out_src.append(f"def {name}({params_as_str}){return_annotations}: ...")
        out_src.append("\n")
    return out_src


def _get_bases(cls, local_attrs, additional_classes) -> list:
    res = []
    for b in cls.__bases__:
        if b is object or b.__module__ == "typing":
            continue
        if not _is_sqlalchemy(b):
            the_type = _get_type_info(b, local_attrs, additional_classes)
            if the_type != "Any":
                res.append(the_type)
    return res


def get_stubs_of_other_classes(
    *, other_classes, local_attrs, additional_classes, additional_imports
):
    out_src = []
    for cls_name, cls in other_classes.items():
        if cls.__module__ != local_attrs["__name__"]:
            if cls_name not in additional_imports:
                out_src += [f"class {cls_name}:", f"{INDENT}pass", ""]
            continue

        bases = _get_bases(cls, local_attrs, additional_classes)
        method_info = _get_methods_info(
            cls, locals_attrs=local_attrs, additional_classes=additional_classes
        )
        bases_str = f"({', '.join(bases)})" if bases else ""
        out_src.append(f"class {cls_name}{bases_str}:")
        if not method_info:
            out_src.append(f"{INDENT}pass")
        out_src.append("")
        out_src += [f"{INDENT}{m}" for m in method_info]
        out_src.append("\n")
    return out_src


def get_typevars(attrs):
    res = [
        f'{k} = TypeVar("{k}")' for k in attrs if isinstance(attrs[k], typing.TypeVar)
    ]
    return [""] + res + [""]


def create_pyi(calling_source_file, attrs: dict, only_current_module: bool = True):
    full_path: Path = Path(calling_source_file)
    pyi_path = (full_path.parent / f"{full_path.stem}.pyi").resolve()
    out_src = []
    additional_imports = []
    imported = list(get_imports(attrs.get("__file__")))
    if only_current_module:
        for level, pkg_name, val_info, alias in imported:
            level_s = "." * level
            val = ".".join(val_info)
            alias = alias or val
            if val and pkg_name:
                if val != "*":
                    out_src.append(f"from {level_s}{pkg_name} import {val} as {alias}")
                    additional_imports.append(alias)
                else:
                    out_src.append(f"from {level_s}{pkg_name} import {val}")
            elif pkg_name:
                out_src.append(f"import {level_s}{pkg_name}")
                additional_imports.append(pkg_name)
            else:
                out_src.append(f"import {level_s}{alias}")
                additional_imports.append(alias)

    enum_classes = _get_enum_classes(attrs, only_calling_module=only_current_module)
    if enum_classes and enum not in additional_imports:
        out_src += ["import enum", ""]
    struct_classes = _get_struct_classes(attrs, only_calling_module=only_current_module)
    other_classes = _get_other_classes(attrs, only_calling_module=only_current_module)
    functions = _get_functions(attrs, only_calling_module=only_current_module)

    out_src += get_typevars(attrs)

    additional_classes = set()
    out_src += _get_consts(
        attrs,
        additional_classes=additional_classes,
        additional_imports=additional_imports,
    )

    out_src += get_stubs_of_enums(
        enum_classes, local_attrs=attrs, additional_classes=additional_classes
    )
    out_src += get_stubs_of_other_classes(
        other_classes=other_classes,
        local_attrs=attrs,
        additional_classes=additional_classes,
        additional_imports=additional_imports,
    )
    out_src += get_stubs_of_structures(
        struct_classes, local_attrs=attrs, additional_classes=additional_classes
    )

    out_src += get_stubs_of_functions(
        functions, local_attrs=attrs, additional_classes=additional_classes
    )

    from_future_import = [
        (i, s) for i, s in enumerate(out_src) if s.startswith("from __future__")
    ]
    for number_of_deletions, (i, s) in enumerate(from_future_import):
        out_src = (
            out_src[: i - number_of_deletions] + out_src[i + 1 - number_of_deletions :]
        )
    out_src = (
        [x[1] for x in from_future_import]
        + add_imports(
            local_attrs=attrs,
            additional_classes=additional_classes,
            existing_imports=set(additional_imports),
        )
        + out_src
    )

    out_src = AUTOGEN_NOTE + out_src
    out_s = "\n".join(out_src)
    with open(pyi_path, "w", encoding="UTF-8") as f:
        f.write(out_s)


def _get_consts(attrs, additional_classes, additional_imports):
    def _is_of_builtin(v) -> bool:
        return isinstance(v, (int, float, str, dict, list, set, complex, bool))

    def _as_builtin(v) -> str:
        return v if isinstance(v, (int, float, str, complex, bool)) else v.__class__()

    res = []
    annotations = attrs.get("__annotations__", None) or {}
    constants = {
        k: v
        for (k, v) in attrs.items()
        if (_is_of_builtin(v) or v is None)
        and not k.startswith("__")
        and k not in additional_imports
    }
    for c in constants:
        the_type = (
            _get_type_info(annotations[c], attrs, additional_classes)
            if c in annotations
            else None
        )
        type_str = f": {the_type}" if the_type else ""
        val = (
            str(wrap_val(_as_builtin(attrs[c])))
            if _is_of_builtin(attrs[c])
            else "None"
            if attrs[c] is None
            else ""
        )
        val_st = f" = {val}" if val else ""
        res.append(f"{c}{type_str}{val_st}")
        res.append("")
    return res


def create_stub_for_file(abs_module_path: str, src_root: str, stubs_root: str = None):
    ext = os.path.splitext(abs_module_path)[-1].lower()
    if ext != ".py":
        return
    stem = Path(abs_module_path).stem
    dir_name = str(Path(abs_module_path).parent)
    relative_dir = relpath(dir_name, src_root)
    package_name = ".".join(Path(relative_dir).parts)
    module_name = stem if stem != "__init__" else package_name
    sys.path.append(str(Path(dir_name).parent))
    spec = importlib.util.spec_from_file_location(module_name, abs_module_path)
    the_module = importlib.util.module_from_spec(spec)
    if not the_module.__package__:
        the_module.__package__ = package_name
    spec.loader.exec_module(the_module)

    pyi_dir = (
        Path(stubs_root) / Path(relative_dir)
        if stubs_root
        else Path(abs_module_path).parent
    )
    pyi_dir.mkdir(parents=True, exist_ok=True)
    (pyi_dir / Path("__init__.pyi")).touch(exist_ok=True)

    pyi_path = (pyi_dir / f"{stem}.pyi").resolve()
    if not getattr(the_module, "__package__", None):
        the_module.__package__ = ".".join(Path(relative_dir).parts)
    create_pyi(str(pyi_path), the_module.__dict__)



def create_pyi_ast(calling_source_file, pyi_path):
    out_src = ["import datetime", "from typing import Optional, Any, Iterable"]
    additional_imports = []
    imported = list(get_imports(calling_source_file))
    found_sqlalchmy = False
    for level, pkg_name, val_info, alias in imported:
        level_s = "." * level
        val = ".".join(val_info)
        alias = alias or val
        if pkg_name and pkg_name.startswith("sqlalchemy"):
            found_sqlalchmy = True
        if val and pkg_name:
            if val != "*":
                out_src.append(f"from {level_s}{pkg_name} import {val} as {alias}")
                additional_imports.append(alias)
            else:
                out_src.append(f"from {level_s}{pkg_name} import {val}")
        elif pkg_name:
            out_src.append(f"import {level_s}{pkg_name}")
            additional_imports.append(pkg_name)
        else:
            out_src.append(f"import {level_s}{alias}")
            additional_imports.append(alias)

    if found_sqlalchmy:
        out_src.extend([""] * 3)
        models = get_models(calling_source_file)
        out_src += models_to_src(models)
    out_src = AUTOGEN_NOTE + out_src
    out_s = "\n".join(out_src)
    with open(pyi_path, "w", encoding="UTF-8") as f:
        f.write(out_s)


def create_stub_for_file_using_ast(
    abs_module_path: str, src_root: str, stubs_root: str = None
):
    ext = os.path.splitext(abs_module_path)[-1].lower()
    if ext != ".py":
        return
    stem = Path(abs_module_path).stem
    dir_name = str(Path(abs_module_path).parent)
    relative_dir = relpath(dir_name, src_root)
    package_name = ".".join(Path(relative_dir).parts)
    module_name = stem if stem != "__init__" else package_name

    pyi_dir = (
        Path(stubs_root) / Path(relative_dir)
        if stubs_root
        else Path(abs_module_path).parent
    )
    pyi_dir.mkdir(parents=True, exist_ok=True)
    (pyi_dir / Path("__init__.pyi")).touch(exist_ok=True)

    pyi_path = (pyi_dir / f"{stem}.pyi").resolve()
    create_pyi_ast(abs_module_path, str(pyi_path))



import ast
import typing

from typedpy import Array, Map, Structure

INDENT = " " * 4


def get_imports(path):
    with open(path) as fh:
        root = ast.parse(fh.read(), path)

    for node in ast.iter_child_nodes(root):
        level = 0
        if isinstance(node, ast.Import):
            module = []
        elif isinstance(node, ast.ImportFrom):
            module = node.module
            level = node.level
        else:
            continue

        for n in node.names:
            yield (level, module, n.name.split("."), n.asname)



class FunctionInfo(Structure):
    name: str
    positional_args: Array[str]
    keyword: typing.Optional[str]
    keyword_only_args: Array[str]
    returns: str
    decorators: Array[str]

    _ignore_none = True


class ModelClass(Structure):
    name: str
    bases: Array[str]
    columns: Map[str, str]
    relationships: Map[str, str]
    functions: Array[FunctionInfo]

    @staticmethod
    def type_by_sqlalchemy_type(sqlalchemy_type: str) -> str:
        the_type = sqlalchemy_type.split(".")[-1]
        return {
            "BigInteger": "int",
            "Integer": "int",
            "Boolean": "bool",
            "String": "str",
            "DateTime": "datetime.datetime",
            "Date": "datetime.date",
            "Float": "float",
            "Interval": "datetime.timedelta",
            "Time": "datetime.time",
        }.get(the_type, "Any")


def _get_param_type(ast_type):
    if isinstance(ast_type, ast.Name):
        return ast_type.id
    elif isinstance(ast_type, ast.Subscript):
        value = ast_type.value.id
        if isinstance(ast_type.slice, ast.Tuple):
            args = [_get_param_type(x) for x in ast_type.slice.dims]
        else:
            args = [ast_type.slice.id]
        return f"{value}[{', '.join(args)}]"


def _extract_kw_args(args_info) -> str:
    if args_info.kwarg:
        return f"**{args_info.kwarg.arg}"
    return ""


def _extract_kwonly_arg(arg, default) -> str:
    the_type = f": {_get_param_type(arg.annotation)}" if arg.annotation else ""
    default_s = " = None" if default is not None else ""
    the_type = f": {_get_param_type(arg.annotation)}" if arg.annotation else ""
    return f"{arg.arg}{the_type}{default_s}"


def _extract_arg(arg):
    the_type = f": {_get_param_type(arg.annotation)}" if arg.annotation else ""
    return f"{arg.arg}{the_type}"


def method_to_src(func: FunctionInfo):
    res = []
    for d in func.decorators:
        res.append(f"{INDENT}@{d}")

    res.append(f"{INDENT}def {func.name}(")
    for a in func.positional_args:
        res.append(f"{INDENT*2}{a},")
    if func.keyword_only_args:
        res.append(f"{INDENT*2}*,")
        for a in func.keyword_only_args:
            res.append(f"{INDENT * 2}{a},")
    if func.keyword:
        res.append(f"{INDENT * 2}{func.keyword},")
    returns_s = f" -> {func.returns}" if func.returns else ""
    res.append(f"{INDENT}){returns_s}: ...")
    res.append("")
    return res




def models_to_src(models: typing.Iterable[ModelClass]) -> typing.Iterable[str]:
    res = []
    for model in models:
        bases = f"({', '.join(model.bases)})" if model.bases else ""
        res.append(f"class {model.name}{bases}:")
        fields_with_type = [f"{INDENT}{field}: {the_type}" for (field, the_type) in model.columns.items()]
        for field in fields_with_type:
            res.append(field)
        for rel in model.relationships:
            res.append(f"{INDENT}{rel}: Any")
        fields_arg_list = [f"{INDENT * 2}{f} = None," for f in fields_with_type]
        res.append(f"{INDENT}def __init__(self,")
        res.extend(fields_arg_list)
        res.append(f"{INDENT}): ...")
        res.append("")
        for func in model.functions:
            res.extend(method_to_src(func))
            res.append("")
        res.append("")

    return res


def get_models(path) -> typing.Iterable[ModelClass]:
    with open(path) as fh:
        root = ast.parse(fh.read(), path)

    res = []
    for node in ast.iter_child_nodes(root):
        if isinstance(node, ast.ClassDef):
            if node.bases:
                bases = [x.id for x in node.bases if isinstance(x, ast.Name)]
                if "Base" in bases:
                    bases.remove("Base")
                    class_name = node.name
                    body = node.body
                    functions = []
                    relationships = {}
                    columns = {}
                    for b in body:
                        if isinstance(b, ast.Assign) and isinstance(b.value, ast.Call):
                            field_name = b.targets[0].id
                            function_name = b.value.func.id
                            if function_name == "Column":
                                args = b.value.args
                                first = args[0]
                                column_type = first.id if isinstance(first, ast.Name) else first.func.id
                                columns[
                                    field_name
                                ] = ModelClass.type_by_sqlalchemy_type(column_type)
                            if function_name == "relationship":
                                args = b.value.args
                                first = args[0]
                                entity_type = first.id if isinstance(first, ast.Name) else first.value
                                relationships[field_name] = entity_type
                        if isinstance(b, ast.FunctionDef):
                            decorators = [d.id for d in b.decorator_list]
                            return_type = (
                                _get_param_type(b.returns) if b.returns else ""
                            )
                            args = [_extract_arg(a) for a in b.args.args]
                            keyword_args = _extract_kw_args(b.args)
                            keyword_only_args = [
                                _extract_kwonly_arg(a, d)
                                for a, d in zip(b.args.kwonlyargs, b.args.kw_defaults)
                            ]
                            functions.append(
                                FunctionInfo(
                                    name=b.name,
                                    positional_args=args,
                                    keyword=keyword_args,
                                    keyword_only_args=keyword_only_args,
                                    returns=return_type,
                                    decorators=decorators,
                                )
                            )
                    res.append(
                        ModelClass(
                            name=class_name,
                            bases=bases,
                            columns=columns,
                            relationships=relationships,
                            functions=functions,
                        )
                    )
    return res




"""
The Skeleton classes to support strictly defined structures:
Structure, Field, StructureReference, ClassReference, TypedField
"""
import enum
import inspect
from builtins import issubclass
from copy import deepcopy
from collections import OrderedDict, deque, defaultdict
from inspect import Signature, Parameter, signature, currentframe
import sys
import typing
import hashlib

from typing import get_type_hints, Iterable

from .commons import raise_errs_if_needed, wrap_val, _is_sunder, _is_dunder
from .utility import type_is_generic

REQUIRED_FIELDS = "_required"
DEFAULTS = "_defaults"
ADDITIONAL_PROPERTIES = "_additionalProperties"
IS_IMMUTABLE = "_immutable"
OPTIONAL_FIELDS = "_optional"
MUST_BE_UNIQUE = "_must_be_unique"
IGNORE_NONE_VALUES = "_ignore_none"
SERIALIZATION_MAPPER = "_serialization_mapper"
DESERIALIZATION_MAPPER = "_deserialization_mapper"
SPECIAL_ATTRIBUTES = {
    REQUIRED_FIELDS,
    ADDITIONAL_PROPERTIES,
    IS_IMMUTABLE,
    DEFAULTS,
    OPTIONAL_FIELDS,
    SERIALIZATION_MAPPER,
    DESERIALIZATION_MAPPER,
    IGNORE_NONE_VALUES,
}

MAX_NUMBER_OF_INSTANCES_TO_VERIFY_UNIQUENESS = 100000

T = typing.TypeVar("T")


class ImmutableMixin:
    """
    Helper for making a field immutable
    """

    _field_definition = None
    _instance = None

    def _get_defensive_copy_if_needed(self, value):
        return deepcopy(value) if self._is_immutable() else value

    def _is_immutable(self):
        if getattr(self._field_definition, "_immutable", False):
            return True
        instance = getattr(self, "_instance", None)
        return getattr(self._instance, IS_IMMUTABLE, False) if instance else False

    def _raise_if_immutable(self):
        if self._is_immutable():
            name = getattr(self, "_name", None)
            raise ValueError(f"{name}: Field is immutable")


def make_signature(
    names, required, additional_properties, bases_params_by_name, bases_required
):
    """
    Make a signature that will be used for the constructor of the Structure
    :param names: names of the properties
    :param required: list of required properties
    :param additional_properties: are additional properties allowed?
    :param bases_params_by_name: parameters taken from base classes (i.e. MRO)
           in case of inheritance
    :return: the signature
    """
    all_names = set(names) | set(bases_params_by_name.keys())
    non_default_args_for_class = OrderedDict(
        [
            (name, Parameter(name, Parameter.POSITIONAL_OR_KEYWORD))
            for name in all_names
            if name in required
        ]
    )
    non_default_args_for_bases = OrderedDict(
        [
            (name, param)
            for (name, param) in bases_params_by_name.items()
            if (name in required or name in bases_required)
        ]
    )
    non_default_args = list(
        {**non_default_args_for_bases, **non_default_args_for_class}.values()
    )

    default_args_for_class = OrderedDict(
        [
            (name, Parameter(name, Parameter.POSITIONAL_OR_KEYWORD, default=None))
            for name in names
            if name not in required
        ]
    )
    default_args_for_bases = OrderedDict(
        [
            (name, param)
            for (name, param) in bases_params_by_name.items()
            if (name not in required and name not in bases_required)
        ]
    )
    default_args = list({**default_args_for_bases, **default_args_for_class}.values())

    additional_args = (
        [Parameter("kwargs", Parameter.VAR_KEYWORD)] if additional_properties else []
    )

    return Signature(non_default_args + default_args + additional_args)


def get_base_info(bases):
    """
    Extract the parameters from all the base classes to support inheritance of Structures.
    :param bases: list of base classes
    :return:  a tuple of: (the parameters from the base classes, a list of the names of
              the required ones)
    """
    bases_params = OrderedDict()
    bases_required = []
    base_structures = (
        [
            base
            for base in bases
            if issubclass(base, Structure) and base is not Structure
        ]
        if "Structure" in globals()
        else []
    )
    for base in base_structures:
        for k, param in getattr(base, "__signature__").parameters.items():
            if k not in bases_params:
                if param.default is not None and param.kind != Parameter.VAR_KEYWORD:
                    bases_required.append(k)
                bases_params[k] = param
        additional_props = base.__dict__.get(ADDITIONAL_PROPERTIES, True)
        if additional_props and bases_params["kwargs"].kind == Parameter.VAR_KEYWORD:
            del bases_params["kwargs"]

    return bases_params, bases_required


def _check_for_final_violations(classes):
    def is_sub_class(c, base):
        return issubclass(c, base) and c != base

    _, *inherited_class = classes
    for c in inherited_class:
        if "FinalStructure" in globals() and isinstance(c, StructMeta):
            if "FinalStructure" in globals() and is_sub_class(c, FinalStructure):
                raise TypeError(
                    f"Tried to extend {c.__name__}, which is a FinalStructure. This is forbidden"
                )
            if "ImmutableStructure" in globals() and is_sub_class(
                c, ImmutableStructure
            ):
                raise TypeError(
                    f"Tried to extend {c.__name__}, which is an ImmutableStructure. This is forbidden"
                )

        if "_FieldMeta" in globals() and isinstance(c, _FieldMeta):
            if "ImmutableField" in globals() and is_sub_class(c, ImmutableField):
                raise TypeError(
                    f"Tried to extend {c.__name__}, which is an ImmutableField. This is forbidden"
                )


def _or_fields(first, other):
    from .fields import AnyOf
    from .enum import Enum

    if isinstance(other, (Field, Structure, _FieldMeta, StructMeta)):
        return AnyOf[first, other]
    if isinstance(other, (str, int, float, bool, list, set, dict, tuple)):
        return AnyOf[first, Enum(values=[other])]
    converted = convert_basic_types(other)
    if converted:
        return AnyOf[first, converted]
    raise TypeError(f"| is Supported only between field types; Got {first} and {other}")


class _FieldMeta(type):
    _registry = {}

    def __new__(cls, name, bases, cls_dict):
        clsobj = super().__new__(cls, name, bases, dict(cls_dict))
        _check_for_final_violations(clsobj.mro())
        return clsobj

    def __or__(self, other):
        return _or_fields(self, other)

    def __getitem__(cls, val):
        if isinstance(val, Field):
            return val
        elif Field in getattr(val, "__mro__", {}):
            return val()
        elif Structure in getattr(val, "__mro__", {}):
            return ClassReference(val)
        elif is_function_returning_field(val):
            return val()
        elif val is None:
            return NoneField()
        else:
            try:
                converted = convert_field_type_if_possible(val)
                if converted is None:
                    raise TypeError
                return _FieldMeta.__getitem__(cls, converted)
            except TypeError:

                def get_state(value):
                    raise TypeError(
                        "pickling of implicit wrappers for non-Typedpy fields are unsupported"
                    )

                if not isinstance(val, type):
                    raise TypeError(
                        f"Unsupported field type in definition: {wrap_val(val)}"
                    )
                the_class = val.__name__
                if the_class in _FieldMeta._registry:
                    return _FieldMeta._registry[the_class]
                short_hash = hashlib.sha256(the_class.encode("utf-8")).hexdigest()[:8]
                new_name = f"Field_{the_class}_{short_hash}"
                class_as_field = create_typed_field(new_name, val)
                class_as_field.__getstate__ = get_state
                _FieldMeta._registry[the_class] = class_as_field
                return class_as_field()


class UniqueMixin:
    def defined_as_unique(self):
        return getattr(self, MUST_BE_UNIQUE, False)

    def __manage_uniqueness__(self):
        myclass = self.__class__
        if (
            getattr(myclass, MUST_BE_UNIQUE, False)
            and len(getattr(myclass, "_ALL_INSTANCES", set()))
            < MAX_NUMBER_OF_INSTANCES_TO_VERIFY_UNIQUENESS
        ):
            hash_of_instance = self.__hash__()
            if hash_of_instance in getattr(myclass, "_ALL_INSTANCES", set()):
                classname = self.__class__.__name__
                raise ValueError(
                    f"Instance copy in {classname}, which is defined as unique. Instance is {self}"
                )
            getattr(myclass, "_ALL_INSTANCES", set()).add(hash_of_instance)

    def __manage_uniqueness_for_field__(self, instance, value):
        if not getattr(instance, "_instantiated", False) or not getattr(
            self, MUST_BE_UNIQUE, False
        ):
            return
        field_name = getattr(self, "_name")
        structure_class_name = instance.__class__.__name__
        all_instances_by_struct_name = getattr(
            self, "_ALL_INSTANCES", defaultdict(dict)
        )
        instance_by_value_for_current_struct = all_instances_by_struct_name[
            structure_class_name
        ]
        if (
            getattr(self, MUST_BE_UNIQUE, False)
            and len(instance_by_value_for_current_struct)
            < MAX_NUMBER_OF_INSTANCES_TO_VERIFY_UNIQUENESS
        ):
            hash_of_field_val = value.__hash__()
            if (
                instance_by_value_for_current_struct.get(hash_of_field_val, instance)
                != instance
            ):
                raise ValueError(
                    f"Instance copy of field {field_name} in {structure_class_name}, which is defined as unique. "
                    f"Instance is {wrap_val(value)}"
                )
            if hash_of_field_val not in instance_by_value_for_current_struct:
                instance_by_value_for_current_struct[hash_of_field_val] = instance


class Field(UniqueMixin, metaclass=_FieldMeta):
    """
    Base class for a field(i.e. property) in a structure.
    Should not be used directly by developers.

    Arguments:
        immutable: optional
            Marks the field as immutable. Typically the developer does not need to use it,
            as there is a high level API for making a field immutable

        is_unique: optional
            Marks a field as unique within this its Structure.
            as there is a high level API for making a field immutable

            .. code-block:: python

                    class SSID(String): pass

                    class Person(Structure):
                            ssid: SSID(is_unique=True)
                            name: String

                    Person(ssid="1234", name="john")
                    # the next line will raise an exception "Instance copy of field ssid in Person"
                    Person(ssid="1234", name="jeff")

            Alternatively, you can use the "@unique" decorator on the class definition of the Field.
            Refer to "Uniqueness" section for more detail.

        default: optional
            default value in case no value was assigned. Setting it makes
            the field implicitly optional.
            Default values are validated based on the field definition like any other value
            assignment.
    """

    def __init__(self, name=None, immutable=None, is_unique=None, default=None):
        self._name = name
        self._default = default
        if is_unique in [True, False]:
            setattr(self, MUST_BE_UNIQUE, is_unique)
            if is_unique:
                self._ALL_INSTANCES = defaultdict(dict)
        if immutable is not None:
            self._immutable = immutable
        if default:
            default_val = default() if callable(default) else default
            self._try_default_value(default_val)

    def __or__(self, other):
        return _or_fields(self, other)

    def _try_default_value(self, default):
        try:
            self._name = self._name or "value"
            self.__set__(Structure(), default)
        except Exception as e:
            raise e.__class__(
                f"Invalid default value: {wrap_val(default)}; Reason: {str(e)}"
            ) from e

    def __get__(self, instance, owner):
        def get_field_with_inheritance(name):
            if name in owner.__dict__:
                return owner.__dict__[name]
            field_by_name = owner.get_all_fields_by_name()
            return field_by_name[name]

        if instance is not None and self._name not in instance.__dict__:
            default_value = (
                self._default() if callable(self._default) else self._default
            )
            return default_value
        res = (
            instance.__dict__[self._name]
            if instance is not None
            else get_field_with_inheritance(self._name)
        )
        is_immutable = (
            instance is not None
            and getattr(instance, IS_IMMUTABLE, False)
            or getattr(self, IS_IMMUTABLE, False)
        )
        needs_defensive_copy = (
            not isinstance(
                res, (ImmutableMixin, int, float, str, bool, enum.Enum, Field)
            )
            or res is None
        )
        return deepcopy(res) if (is_immutable and needs_defensive_copy) else res

    def __set__(self, instance, value):
        if getattr(self, IS_IMMUTABLE, False) and self._name in instance.__dict__:
            raise ValueError(f"{self._name}: Field is immutable")
        if getattr(self, IS_IMMUTABLE, False) and not getattr(
            self, "_custom_deep_copy_implementation", False
        ):
            needs_defensive_copy = (
                not isinstance(
                    value,
                    (
                        ImmutableMixin,
                        int,
                        float,
                        str,
                        bool,
                        enum.Enum,
                        Field,
                        ImmutableStructure,
                    ),
                )
                or value is None
            )
            try:
                instance.__dict__[self._name] = (
                    deepcopy(value) if needs_defensive_copy else value
                )
            except TypeError:
                raise TypeError(
                    f"{self._name} cannot be immutable, as its type does not support pickle."
                )
        else:
            self.__manage_uniqueness_for_field__(instance, value)
            instance.__dict__[self._name] = value
            instance.__manage__uniqueness_of_all_fields__()
        if getattr(instance, "_instantiated", False) and not getattr(
            instance, "_skip_validation", False
        ):
            instance.__validate__()

    def __serialize__(self, value):  # pylint: disable=R0201
        return value

    def __str__(self):
        def as_str(the_val):
            """
            convert to string or a list of strings
            :param the_val: a Field or a list of Fields
            :return: a string representation
            """
            if hasattr(the_val, "__iter__"):
                return f"[{', '.join([str(v) for v in the_val])}]"
            return str(the_val)

        name = self.__class__.__name__
        props = []
        for k, val in sorted(self.__dict__.items()):
            if val is not None and not k.startswith("_"):
                strv = f"'{val}'" if isinstance(val, str) else as_str(val)
                props.append(f"{k} = {strv}")

        propst = f". Properties: {', '.join(props)}" if props else ""
        return f"<{name}{propst}>"

    def _set_immutable(self, immutable: bool):
        self._immutable = immutable

    @property
    def get_type(self):
        return typing.Any


class TypedField(Field):
    """
    A strictly typed base field.
    Should not be used directly. Instead, use :func:`create_typed_field`
    """

    _ty = object

    def _validate(self, value):
        def err_prefix():
            return f"{self._name}: " if self._name else ""

        if not isinstance(value, self._ty):
            raise TypeError(f"{err_prefix()}Expected {self._ty}; Got {wrap_val(value)}")

    def __set__(self, instance, value):
        if not getattr(instance, "_skip_validation", False):
            self._validate(value)
        super().__set__(instance, value)

    @property
    def get_type(self):
        return self.__class__._ty


# noinspection PyBroadException
def is_function_returning_field(field_definition_candidate):
    python_ver_higher_than_36 = sys.version_info[0:2] != (3, 6)
    if callable(field_definition_candidate) and python_ver_higher_than_36:
        try:
            if len(signature(field_definition_candidate).parameters) > 0:
                raise TypeError("function not allowed to accept any parameters")
            return_value = get_type_hints(field_definition_candidate).get(
                "return", None
            )
            return return_value == Field or Field in getattr(
                return_value.__args__[0], "__mro__", []
            )
        except Exception:
            return False
    return False


def _get_all_fields_by_name(cls):
    all_classes = reversed([c for c in cls.mro() if isinstance(c, StructMeta)])
    all_fields_by_name = {}
    for the_class in all_classes:
        if isinstance(the_class, StructMeta):
            field_names = getattr(the_class, "_fields", [])
            field_by_name = dict([(k, getattr(the_class, k)) for k in field_names])
            all_fields_by_name.update(field_by_name)
    return all_fields_by_name


def _get_all_values_of_attribute(cls, attr_name: str):
    all_classes = reversed([c for c in cls.mro() if isinstance(c, StructMeta)])
    all_values = []
    for the_class in all_classes:
        if issubclass(the_class, Structure):
            attr = getattr(the_class, attr_name, None)
            if isinstance(attr, list):
                all_values.extend(attr)
            elif attr is not None:
                all_values.append(attr)
    return all_values


def _instantiate_fields_if_needed(cls_dict: dict, defaults: dict):
    for key, val in cls_dict.items():
        if (
            key not in SPECIAL_ATTRIBUTES
            and not isinstance(val, Field)
            and not key.startswith("__")
            and (
                Field in getattr(val, "__mro__", []) or is_function_returning_field(val)
            )
        ):
            new_val = val(default=defaults[key]) if key in defaults else val()
            cls_dict[key] = new_val


def _apply_default_and_update_required_not_to_include_fields_with_defaults(
    cls_dict: dict, defaults: dict, fields: list
):
    required_fields = set(cls_dict.get(REQUIRED_FIELDS, []))
    optional_fields = set(cls_dict.get(OPTIONAL_FIELDS, []))
    required_fields_predefined = REQUIRED_FIELDS in cls_dict
    for field_name in fields:
        if field_name in defaults and not getattr(
            cls_dict[field_name], "_default", None
        ):
            if isinstance(defaults[field_name], (list, dict, set)):
                raise ValueError(
                    f"{field_name}: Got a mutable value as default. This is a bug. "
                    "Instead, use a generating function."
                )
            default_value = (
                defaults[field_name]()
                if callable(defaults[field_name])
                else defaults[field_name]
            )
            cls_dict[field_name]._try_default_value(default_value)
            cls_dict[field_name]._default = defaults[field_name]
        if getattr(cls_dict[field_name], "_default", None) is not None:
            if field_name in required_fields:
                required_fields.remove(field_name)
        elif not required_fields_predefined:
            if field_name not in optional_fields:
                required_fields.add(field_name)
    cls_dict[REQUIRED_FIELDS] = list(required_fields)


class StructMeta(type):
    """
    Metaclass for Structure. Manipulates it to ensure the fields are set up correctly.
    """

    @classmethod
    def __prepare__(cls, name, bases):
        return OrderedDict()

    def __new__(cls, name, bases, cls_dict):
        bases_params, bases_required = get_base_info(bases)
        add_annotations_to_class_dict(cls_dict, previous_frame=currentframe().f_back)
        defaults = cls_dict[DEFAULTS]
        _instantiate_fields_if_needed(cls_dict=cls_dict, defaults=defaults)

        for key, val in cls_dict.items():
            if isinstance(val, StructMeta) and not isinstance(val, Field):
                cls_dict[key] = ClassReference(val)
        fields = [key for key, val in cls_dict.items() if isinstance(val, Field)]
        for field_name in fields:
            if field_name.startswith("_") or field_name == "kwargs":
                raise ValueError(f"{field_name}: invalid field name")
            setattr(cls_dict[field_name], "_name", field_name)

        for key, val in cls_dict.items():
            if (
                not any([_is_sunder(key), _is_dunder(key), isinstance(val, Field)])
                and (isinstance(val, type) or type_is_generic(val))
                and Structure.is_non_typedpy_field_assignment_blocked()
            ):
                raise TypeError(f"{key}: assigned a non-Typedpy type: {val}")
        _apply_default_and_update_required_not_to_include_fields_with_defaults(
            cls_dict=cls_dict, defaults=defaults, fields=fields
        )

        cls_dict.pop(DEFAULTS, None)
        clsobj = super().__new__(cls, name, bases, dict(cls_dict))
        _check_for_final_violations(clsobj.mro())
        clsobj._fields = fields

        if hasattr(clsobj, "__annotations__"):
            for key, val in _get_all_fields_by_name(clsobj).items():
                if key not in clsobj.__annotations__ and isinstance(val, TypedField):
                    clsobj.__annotations__[key] = getattr(val, "_ty")

        default_required = (
            list(set(bases_required + fields)) if bases_params else fields
        )
        required = cls_dict.get(REQUIRED_FIELDS, default_required)
        setattr(clsobj, REQUIRED_FIELDS, list(set(bases_required + required)))
        optional_fields = cls_dict.get(OPTIONAL_FIELDS, [])
        for f in optional_fields:
            if f in required or f in bases_required:
                raise ValueError(
                    "optional cannot override prior required in the class or in a base class"
                )
        additional_props = cls_dict.get(ADDITIONAL_PROPERTIES, True)

        sig = make_signature(
            clsobj._fields,
            required,
            additional_props,
            bases_params,
            bases_required=bases_required,
        )
        setattr(clsobj, "__signature__", sig)
        setattr(clsobj, "_field_by_name", _get_all_fields_by_name(clsobj))
        return clsobj

    def __str__(cls):
        name = cls.__name__
        props = []
        for k, val in sorted(cls.__dict__.items()):
            if val is not None and not k.startswith("_"):
                strv = f"'{val}'" if isinstance(val, str) else str(val)
                props.append(f"{k} = {strv}")
        props_list = ", ".join(props)
        return f"<Structure: {name}. Properties: {props_list}>"


def convert_basic_types(v):
    from .fields import (
        Integer,
        Float,
        String,
        Map,
        Array,
        Tuple,
        Set,
        Boolean,
        ImmutableSet,
        AnyOf,
        Anything,
        Deque,
    )

    type_mapping = {
        deque: Deque,
        int: Integer,
        str: String,
        float: Float,
        dict: Map,
        set: Set,
        list: Array,
        tuple: Tuple,
        bool: Boolean,
        frozenset: ImmutableSet,
        typing.Union: AnyOf,
        typing.Any: Anything,
    }
    return type_mapping.get(v, None)


def get_typing_lib_info(v):
    from .fields import AnyOf

    if v == type(None):
        return NoneField()
    if isinstance(v, Field):
        return v
    if inspect.isclass(v) and issubclass(v, Field):
        return v()

    if isinstance(v, StructMeta) and not isinstance(v, Field):
        return ClassReference(v)

    if not type_is_generic(v):
        return convert_basic_types(v)
    origin = getattr(v, "__origin__", None)
    mapped_type = convert_basic_types(origin)
    if mapped_type is None:
        raise TypeError(f"{v} type is not supported")
    args_raw = getattr(v, "__args__", None)
    if not args_raw:
        return mapped_type()
    mapped_args = [
        get_typing_lib_info(a) for a in args_raw if not isinstance(a, typing.TypeVar)
    ]
    if not all(mapped_args):
        if mapped_type == AnyOf:
            for i, arg in enumerate(mapped_args):
                if arg is None:
                    if isinstance(args_raw[i], type):
                        mapped_args[i] = Field[args_raw[i]]
                    else:
                        raise TypeError(f"invalid type {v}")
        else:
            raise TypeError(f"invalid type {v}")
    if mapped_args:
        if mapped_type == AnyOf:
            return mapped_type(fields=mapped_args)
        else:
            mapped_args = mapped_args if len(mapped_args) > 1 else mapped_args[0]
            return mapped_type(items=mapped_args)
    return mapped_type()


def add_annotations_to_class_dict(cls_dict, previous_frame):
    annotations = cls_dict.get("__annotations__", {})
    defaults = {}
    optional_fields = set(cls_dict.get(OPTIONAL_FIELDS, []))
    if isinstance(annotations, dict):
        for k, v in annotations.items():
            if isinstance(v, str) and len(v) < 50:
                # The evil eval is to accommodate "from __future__ import annotations".
                v = eval(
                    v,
                    sys.modules[cls_dict["__module__"]].__dict__,
                    previous_frame.f_locals,
                )
            first_arg = getattr(v, "__args__", [0])[0]
            mros = getattr(first_arg, "__mro__", getattr(v, "__mro__", []))
            if not type_is_generic(v) and (
                isinstance(v, (Field, Structure))
                or Field in mros
                or Structure in mros
                or is_function_returning_field(v)
            ):
                if k in cls_dict:
                    defaults[k] = cls_dict[k]

                cls_dict[k] = v
            else:
                the_type = get_typing_lib_info(v)
                if the_type:
                    from .fields import AnyOf

                    if isinstance(the_type, AnyOf) and getattr(
                        the_type, "_is_optional", False
                    ):
                        optional_fields.add(k)
                    if k in cls_dict:
                        default = cls_dict[k]
                        default_value = default() if callable(default) else default
                        try:
                            if isinstance(the_type, Field):
                                the_type._try_default_value(default_value)
                            else:
                                the_type = the_type(default=default_value)
                        except Exception as e:
                            raise e.__class__(f"{k}: {str(e)}") from e
                        defaults[k] = cls_dict[k]
                    cls_dict[k] = the_type
    if optional_fields:
        cls_dict[OPTIONAL_FIELDS] = optional_fields
    cls_dict[DEFAULTS] = defaults


def convert_field_type_if_possible(the_field):
    first_arg = getattr(the_field, "__args__", [0])[0]
    mros = getattr(first_arg, "__mro__", getattr(the_field, "__mro__", []))
    if not type_is_generic(the_field) and (
        isinstance(the_field, (Field, Structure)) or Field in mros or Structure in mros
    ):
        return the_field
    else:
        return get_typing_lib_info(the_field)


class Structure(UniqueMixin, metaclass=StructMeta):
    """
    The base class to support strictly defined structures. When creating a new instance of
    a Structure, fields must be provided by name.
    Supports basic constructs: string conversion, quality, copy, deep-copy, hash etc.

    Arguments:
        _required: optional
            An array of the mandatory fields. The default is all the fields in the class.
            Example:

            .. code-block:: python

                class Foo(Structure):
                    _required = ['id']

                    id = Integer
                    name = String

                # this is valid:
                Foo(id = 1)

                # this raises an exception:
                Foo(name="John")

        _optional: optional
            If we don't state the "_required" field, we can state which fields are optional instead.
            Example:

            .. code-block:: python

                class Foo(Structure):
                    id = Integer
                    name = String
                    _optional = ['name']


        _additionalProperties(bool): optional
            Is it allowed to add additional properties that are not defined in the class definition?
            the default is True.
            Example:

            .. code-block:: python

                class Foo(Structure):
                    _additionalProperties = False

                    id = Integer

                # this is valid:
                Foo(id = 1)

                # this raises an exception:
                Foo(id = 1, a = 2)

        _ignore_none(bool): optional
             Ignore assignment to None for any field value.
             Default is False.
             Required fields never ignore None (since they are required)

        _serialization_mapper(dict or mapper): optional
             mapper for the purpose of serialization/deserialization
             if no _deserialization_mapper is defined, it is also
             used for deserialization.
             Example:

             .. code-block:: python

                class Foo(Structure):
                    i: int
                    _serialization_mapper = {"i": "j"}

                class Bar(Foo):
                    a: Array
                    _serialization_mapper = mappers.TO_LOWERCASE

                assert Deserializer(Bar).deserialize(
                  {"J": 5, "A": [1, 2, 3]}, keep_undefined=False
                ) == Bar(i=5, a=[1, 2, 3])

        _deserializatin_mapper(dict or mapper): optional
            mapper specifically for deserialization, in case you need to differentiate between
            serialization and deserialization mappers.

    Decorating it with @unique ensures that no all instances of this structure will be unique. It
    will raise an exception otherwise (see "Uniqueness" section).
    """

    _fields = []
    _fail_fast = True

    def __init__(self, *args, **kwargs):
        try:
            bound = getattr(self, "__signature__").bind(*args, **kwargs)
        except TypeError as ex:
            raise TypeError(f"{self.__class__.__name__}: {ex}")
        if "kwargs" in bound.arguments:
            for name, val in bound.arguments["kwargs"].items():
                setattr(self, name, val)
            del bound.arguments["kwargs"]

        field_by_name = self.get_all_fields_by_name()
        defaults_fields = [
            key
            for key, value in field_by_name.items()
            if getattr(value, "_default", None) is not None
            and key not in bound.arguments
        ]
        for field_name in defaults_fields:
            default = getattr(field_by_name[field_name], "_default")
            default_value = default() if callable(default) else default
            setattr(self, field_name, default_value)

        if Structure.failing_fast():
            for name, val in bound.arguments.items():
                setattr(self, name, val)
        else:
            errors = []
            for name, val in bound.arguments.items():
                try:
                    setattr(self, name, val)
                except (TypeError, ValueError) as ex:
                    errors.append(ex)
            raise_errs_if_needed(errors)

        self.__validate__()
        self._instantiated = True
        self.__manage_uniqueness__()
        self.__manage__uniqueness_of_all_fields__()

    def __manage__uniqueness_of_all_fields__(self):
        fields_by_name = self.__class__.get_all_fields_by_name()
        for name, field in fields_by_name.items():
            if field.defined_as_unique():
                field.__manage_uniqueness_for_field__(self, getattr(self, name, None))

    def __setattr__(self, key, value):
        if getattr(self, IS_IMMUTABLE, False):
            if getattr(self, "_instantiated", False):
                raise ValueError(f"{self.__class__.__name__}: Structure is immutable")
            if not getattr(value, IS_IMMUTABLE, False):
                needs_defensive_copy = not isinstance(
                    value,
                    (
                        ImmutableMixin,
                        int,
                        float,
                        str,
                        bool,
                        enum.Enum,
                        Field,
                        ImmutableStructure,
                    ),
                )
                value = deepcopy(value) if needs_defensive_copy else value
        if not any(
            [
                getattr(self, ADDITIONAL_PROPERTIES, True),
                key in self.get_all_fields_by_name(),
                _is_sunder(key),
                _is_dunder(key),
            ]
        ):
            raise ValueError(
                f"{self.__class__.__name__}: trying to set a non-field '{key}' is not allowed"
            )
        if all(
            [
                getattr(self, IGNORE_NONE_VALUES, False),
                value is None,
                key not in getattr(self.__class__, REQUIRED_FIELDS, []),
            ]
        ):
            return

        super().__setattr__(key, value)

        if (
            getattr(self, "_instantiated", False)
            and not _is_dunder(key)
            and not _is_sunder(key)
        ):
            self.__manage_uniqueness__()

    def __getstate__(self):
        fields_by_name = _get_all_fields_by_name(self.__class__)
        return {
            name: field.__serialize__(getattr(self, name, None))
            for (name, field) in fields_by_name.items()
            if name in self.__dict__
        }

    def __str__(self):
        def list_to_str(values):
            as_strings = [to_str(v) for v in values]
            return ",".join(as_strings)

        def dict_to_str(values):
            as_strings = [f"{to_str(k)} = {to_str(v)}" for (k, v) in values.items()]
            return ",".join(as_strings)

        def to_str(the_val):
            if isinstance(the_val, list):
                return f"[{list_to_str(the_val)}]"
            if isinstance(the_val, tuple):
                return f"({list_to_str(the_val)})"
            if isinstance(the_val, set):
                return f"{{{list_to_str(the_val)}}}"
            if isinstance(the_val, dict):
                return f"{{{dict_to_str(the_val)}}}"
            return str(the_val)

        name = self.__class__.__name__
        if name.startswith("StructureReference_") and self.__class__.__bases__ == (
            Structure,
        ):
            name = "Structure"
        props = []
        internal_props = ["_instantiated"]
        for k, val in sorted(self.__dict__.items()):
            if k not in internal_props:
                strv = f"'{val}'" if isinstance(val, str) else to_str(val)
                props.append(f"{k} = {strv}")
        props_list = ", ".join(props)
        return f"<Instance of {name}. Properties: {props_list}>"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        internal_props = ["_instantiated"]
        for k, val in sorted(self.__dict__.items()):
            if k not in internal_props and val != other.__dict__.get(k):
                return False
        for k, val in sorted(other.__dict__.items()):
            if k not in internal_props and val != self.__dict__.get(k):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return str(self).__hash__()

    def __delitem__(self, key):
        if isinstance(getattr(self, REQUIRED_FIELDS), list) and key in getattr(
            self, REQUIRED_FIELDS
        ):
            raise ValueError(f"{key} is mandatory")
        del self.__dict__[key]

    def __validate__(self):
        pass

    def __deepcopy__(self, memo):
        if isinstance(self, ImmutableStructure) and getattr(self, IS_IMMUTABLE, False):
            return self
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        result._skip_validation = True  # pylint: disable=attribute-defined-outside-init
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        delattr(result, "_skip_validation")
        return result

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __dir__(self) -> Iterable[str]:
        internal_props = ["_instantiated"]
        return [k for k in sorted(self.__dict__) if k not in internal_props]

    def __bool__(self):
        internal_props = ["_instantiated"]
        return any(
            v is not None for k, v in self.__dict__.items() if k not in internal_props
        )

    def _additional_serialization(self) -> dict:
        """
        :return: additional fields when serializing a structure,
            Each key is a key in the output json, and each value
            can either be a function with no parameters, a method, or a
            simple value
        """
        if callable(getattr(super(), "_additional_serialization", None)):
            return super()._additional_serialization()
        return {}

    @classmethod
    def get_all_fields_by_name(cls):
        return getattr(cls, "_field_by_name")

    @classmethod
    def get_aggregated_serialization_mapper(cls):
        return _get_all_values_of_attribute(cls, SERIALIZATION_MAPPER)

    @classmethod
    def get_aggregated_deserialization_mapper(cls):
        all_classes = reversed([c for c in cls.mro() if isinstance(c, StructMeta)])
        all_values = []
        for the_class in all_classes:
            if issubclass(the_class, Structure):
                deserialization_mapper = getattr(
                    the_class, DESERIALIZATION_MAPPER, None
                )
                attr = (
                    deserialization_mapper
                    if deserialization_mapper is not None
                    else getattr(the_class, SERIALIZATION_MAPPER, None)
                )
                if isinstance(attr, list):
                    all_values.extend(attr)
                elif attr is not None:
                    all_values.append(attr)
        return all_values

    def _is_wrapper(self):
        field_by_name = _get_all_fields_by_name(self.__class__)
        field_names = list(field_by_name.keys())
        props = self.__class__.__dict__
        required = props.get(REQUIRED_FIELDS, field_names)
        additional_props = props.get(ADDITIONAL_PROPERTIES, True)
        return (
            len(field_names) == 1
            and required == field_names
            and additional_props is False
        )

    def __contains__(self, item):
        if self._is_wrapper():
            field_by_name = _get_all_fields_by_name(self.__class__)
            field_names = list(field_by_name.keys())
            return item in getattr(self, field_names[0], {})

        raise TypeError(f"{self.__class__.__name__} does not support this operator")

    def __iter__(self):
        field_by_name = _get_all_fields_by_name(self.__class__)
        field_names = list(field_by_name.keys())
        val = getattr(self, field_names[0], {})

        if self._is_wrapper() and hasattr(val, "__iter__"):
            return iter(val)
        raise TypeError(f"{self.__class__.__name__} is not a wrapper of an iterable")

    def shallow_clone_with_overrides(self, **kw):
        fields_names = self.get_all_fields_by_name().keys()
        field_value_by_name = {
            f: getattr(self, f) for f in fields_names if getattr(self, f) is not None
        }
        kw_args = {**field_value_by_name, **kw}
        return self.__class__(**kw_args)

    def cast_to(self, cls: type(T)) -> T:
        """
        Shallow copy of the structure as the given class, which should be a subclass
        or superclass of the structure's class
        :param cls: the target class
        :return: an instance of cls
        """
        if (
            issubclass(cls, self.__class__)
            or cls is self.__class__
            or isinstance(self, cls)
        ) and issubclass(cls, Structure):
            that = (
                deepcopy(self)
                if (
                    issubclass(cls, ImmutableStructure)
                    and not issubclass(self.__class__, ImmutableStructure)
                    or issubclass(self.__class__, ImmutableStructure)
                    and not issubclass(cls, ImmutableStructure)
                )
                else self
            )

            fields_names = cls.get_all_fields_by_name().keys()
            field_value_by_name = {
                f: getattr(that, f)
                for f in fields_names
                if getattr(that, f, None) is not None
            }
            return cls(**field_value_by_name)

        raise TypeError(f"cls must be subclass of {self.__class__.__name__}")

    def to_other_class(self, target_class, *, ignore_props=None, **kw) -> T:
        """
        Shallow copy of the fields in the structure and instantiate an instance of the given target_class
        Arguments:
            target_class:
                The target class of the new object to be instantiated from this structure.
                This does not need to be a class:`Structure`.

                Example:

                .. code-block:: python
                    class Person:
                        def __init__(self ,* ,name, age):
                            ...

                    class Foo(Structure):
                        id = Integer
                        name = String

                    person = Foo(id=1, name="john").to_other_class(Person, ignore_props=["id"], age=40)
                    assert person.age == 40
                    assert.person.name == "john"

            ignore_props(list[str]): optional
                a list of field names to be ignored (not copied).

            kw: optional
                any overrides of attributes. For example: "age="40" in the code snippet above

        Returns:
            A new instance of the provided target_class with all the attributes of the current structure,
            except the ones state in the ignore_props parameters, and also the attributes overrides given in
            the keyword arguments.
            :param ignore_props:
        """
        ignore_props = ignore_props if ignore_props else []
        args_from_structure = {
            k: getattr(self, k, None)
            for k in self.get_all_fields_by_name()
            if k not in ignore_props
        }
        kwargs = {**args_from_structure, **kw}
        return target_class(**kwargs)

    @classmethod
    def from_other_class(cls, source_object, *, ignore_props=None, **kw):
        """
        Return a new instance of the current :class:`Structure`, with the attributes of the source_object.
        The optional parameters allow to ignore/override attributes.
        For example:

        .. code-block:: python
                    class PersonModel:
                        def __init__(self ,* ,first_name, age):
                            ...

                    class Person(Structure):
                        id = Integer
                        name = String
                        age = Integer

                    person_model = PersonModel(first_name="john", age=40)
                    person = Person.from_other_class(
                        person_model,
                        id=123,
                        name=person_model.first_name
                    )
                    assert person == Person(name="john", id=123, age=40)

        Arguments:
            source_object:
                The source object to be copied from. Can be of any type.

            ignore_props(list[str]): optional
                The field names to ignore (not copy)

            kw: optional
                explicit overrides/additional field mapping. In the snippet above we
                set the "id" and "name" fields directly.

        Returns:
            The new instance of the current structure type, with the fields set.
        """
        ignore_props = ignore_props if ignore_props else []
        args_from_model = {
            k: getattr(source_object, k, None)
            for k in cls.get_all_fields_by_name()
            if k not in ignore_props
        }
        kwargs = {**args_from_model, **kw}
        return cls(**kwargs)

    @staticmethod
    def set_fail_fast(fast_fail: bool):
        Structure._fail_fast = fast_fail

    @staticmethod
    def failing_fast():
        return Structure._fail_fast

    @staticmethod
    def set_block_non_typedpy_field_assignment(flag=True):
        Structure._block_non_typedpy_field_assignment = flag

    @staticmethod
    def is_non_typedpy_field_assignment_blocked():
        return getattr(Structure, "_block_non_typedpy_field_assignment", True)

    @classmethod
    def omit(cls, *fields_to_omit, class_name: str = ""):
        """
        Define a new Structure class with all the fields of the given class, except for the omitted ones.
         For Example:

         .. code-block:: python

             class Foo(ImmutableStructure):
                i: int
                d: dict[str, int] = dict
                s: set
                a: str
                b: Integer

             class Bar(Foo.omit("a", "b")):
                x: int

        "Bar" has the fields: i, d, s, x. Note that Bar does not extend Foo, but it is a Structure class.
         It does copy attributes like serialization mappers, _ignore_none,
        but Bar can override any of them.

        Another valid usage:

        .. code-block:: python

            Bar = Foo.omit("a", "b", "i", "s")
            bar = Bar(d={"a": 5})

        """

        cls_dict = _init_class_dict(cls)
        cls_dict[REQUIRED_FIELDS] = [
            x for x in getattr(cls, REQUIRED_FIELDS) if x not in fields_to_omit
        ]
        for k, v in cls.get_all_fields_by_name().items():
            if k not in fields_to_omit:
                cls_dict[k] = v

        classname = class_name if class_name else f"Omit{cls.__name__}"
        newclass = type(classname, (Structure,), cls_dict)

        return newclass

    @classmethod
    def pick(cls, *fields_to_pick, class_name: str = ""):
        """
        Define a new Structure class with that picks specific fields from a predefined class.
         For Example:

         .. code-block:: python

             class Foo(ImmutableStructure):
                i: int
                d: dict[str, int] = dict
                s: set
                a: str
                b: Integer

             class Bar(Foo.pick("a", "b")):
                x: int

        "Bar" has the fields: a, b, x. Note that Bar does not extend Foo, but it is a Structure class.
         It does copy attributes like serialization mappers, _ignore_none,
        but Bar can override any of them.

        Another valid usage:

        .. code-block:: python

            Bar = Foo.pick("d")
            bar = Bar(d={"a": 5})

        """
        cls_dict = _init_class_dict(cls)

        for k, v in cls.get_all_fields_by_name().items():
            if k in fields_to_pick:
                cls_dict[k] = v
        cls_dict[REQUIRED_FIELDS] = [
            x for x in getattr(cls, REQUIRED_FIELDS) if x in fields_to_pick
        ]

        classname = class_name if class_name else f"Pick{cls.__name__}"
        newclass = type(classname, (Structure,), cls_dict)

        return newclass


def _init_class_dict(cls):
    attributes_to_include = {
        "_fields",
        IGNORE_NONE_VALUES,
        DEFAULTS,
    }
    cls_dict = {}
    for k, v in cls.__dict__.items():
        if k in attributes_to_include:
            cls_dict[k] = v

    return cls_dict


class FinalStructure(Structure):
    pass


def unique(cls):
    if issubclass(cls, Structure):
        setattr(cls, MUST_BE_UNIQUE, True)
        cls._ALL_INSTANCES = set()
    elif issubclass(cls, Field):
        setattr(cls, MUST_BE_UNIQUE, True)
        cls._ALL_INSTANCES = defaultdict(dict)
    return cls


class ImmutableStructure(Structure):
    """
    A base class for a structure in which non of the fields can be updated post-creation
    Example:

    .. code-block:: python

        class B(ImmutableStructure):
            _required = []
            y = Number
            z = Array[Number]
            m = Map[String, Number]

        b = B(y = 3, z = [1,2,3], m = {'a': 1, 'b': 2})

        # any of the following lines will raise an exception:
        b.y = 1
        b.z[1] += 1
        b.m['c'] = 4
        b.z.clear()
        b.m.pop('a')

    ImmutableStructure class (as the class B in the example above) are not allowed to be extended.
    This is to ensure any instance of ImmutableStructure is indeed immutable.
    """

    _immutable = True


class NoneField(TypedField):
    """
    A field that maps to a single allowable value: None.
    By default, fields cannot be assigned None (i.e. "Null Safety"). NoneField allows to do so.
    This is useful to define optional fields or optional values such as:

    .. code-block:: python

       class Foo(Structure):
           optional_1: typing.Optional[Array]     # NoneField is used implicitly
           optional_2: AnyOf[Array, NoneField]
           optional_3: AnyOf[Array, None]         # the conversion from None to NoneField is implicit

           arr_maybe_int_1: Array[AnyOf[Integer, NoneField]]
           arr_maybe_int_2: Array[AnyOf[Integer, None]]    # the conversion from None to NoneField is implicit

    """

    _ty = type(None)


class ValidatedTypedField(TypedField):
    def __set__(self, instance, value):
        self._validate_func(value)  # pylint: disable=E1101
        super().__set__(instance, value)


def create_typed_field(classname, cls, validate_func=None):
    """
    Factory that generates a new class for a :class:`Field` as a wrapper of any class.
    Example:
    Given a class Foo, and a validation function for the value in Foo - validate_foo, the line

    .. code-block:: python

        ValidatedFooField = create_typed_field("FooField", Foo, validate_func=validate_foo)

    Generates a new :class:`Field` class that validates the content using validate_foo, and can be
    used just like any :class:`Field` type.

    .. code-block:: python

        class A(Structure):
            foo = ValidatedFooField
            bar = Integer

        # asumming we have an instance of Foo, called my_foo:
        A(bar=4, foo=my_foo)

    Arguments:

        classname(`str`):
            the content must not match any of the fields in the lists
    """

    def validate_wrapper(cls, value):
        if validate_func is None:
            return
        validate_func(value)

    return type(
        classname,
        (ValidatedTypedField,),
        {"_validate_func": validate_wrapper, "_ty": cls},
    )


class ClassReference(TypedField):
    """
    A field that is a reference to another Structure instance.
    """

    def __init__(self, cls):
        self._ty = cls
        super().__init__(cls)

    def __str__(self):
        return f"<ClassReference: {self._ty.__name__}>"

    @property
    def get_type(self):
        return self._ty


class ImmutableField(Field):
    """
    A mixin that makes a field class immutable.
    For Example:

     .. code-block:: python

         class MyFieldType(Field): .....

         class MyImmutableFieldType(ImmutableField, MyFieldType): pass

         # that's all you have to do to make MyImmutableFieldType immutable.

    ImmutableField class (as the class MyImmutableFieldType in the example above) are
    not allowed to be extended. This is to ensure any instance of
    ImmutableField is indeed immutable.
    """

    _immutable = True


class AbstractStructure(Structure):
    """
    Defines a Structure class that cannot be instantiated because it is abstract.
     To instantiate, you are required to extend it.
    """

    def __init__(self, *args, **kwargs):
        found = False
        for clz in self.__class__.__bases__:
            if clz is AbstractStructure:
                found = True
                break
        if found:
            raise TypeError("Not allowed to instantiate an abstract Structure")
        super().__init__(*args, **kwargs)



import argparse
import fnmatch
from pathlib import Path

from typedpy.type_helpers import create_stub_for_file, create_stub_for_file_using_ast


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("src_root_dir", help="source root directory")
    parser.add_argument(
        "src_script_path", help="absolute path of python script to process"
    )
    parser.add_argument(
        "-s",
        "--stubs-dir",
        type=str,
        default=".stubs",
        help="source directory of stubs. Default is .stubs",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        type=str,
        help="exclude patterns in the form path1:path2:path3",
    )
    parser.add_argument('--ast',
                        action=argparse.BooleanOptionalAction,
                        help="use AST instead of reflection(currently implemented only for sqlalchemy ORM)")

    args = parser.parse_args()

    src_root_abs_path = str(Path(args.src_root_dir).resolve())
    input_file_abs_path = str(Path(args.src_script_path).resolve())
    stub_dir_abs_path = str(Path(src_root_abs_path) / Path(args.stubs_dir))
    use_ast = args.ast

    exclude = args.exclude.split(":") if args.exclude else []
    for x in exclude:
        if fnmatch.fnmatch(input_file_abs_path, x):
            return

    func = create_stub_for_file_using_ast if use_ast else create_stub_for_file
    func(input_file_abs_path, src_root_abs_path, stub_dir_abs_path)


if __name__ == "__main__":
    main()



"""
Definitions of various types of fields. Supports JSON draft4 types.
"""
import re
import typing
from collections import OrderedDict, deque
from collections.abc import Iterable

from copy import deepcopy
from functools import reduce
from decimal import Decimal, InvalidOperation

from .commons import python_ver_atleast_39, wrap_val
from .structures import (
    Field,
    Structure,
    TypedField,
    ClassReference,
    StructMeta,
    ImmutableMixin,
    _FieldMeta,
    NoneField,
    ImmutableField,
)


class SerializableField(Field):
    """
    An abstract class for a field that has custom serialization or deserialization.
    can override the method.

    .. code-block:: python

      serialize(self, value),
      deserialize(self, value)

    These methods are not being used for pickling.
    """

    def serialize(self, value):  # pylint: disable=no-self-use
        return value

    def deserialize(self, value):  # pylint: disable=no-self-use
        return value


def _map_to_field(item):
    item = item[0] if isinstance(item, (list, tuple)) and len(item) == 1 else item
    if isinstance(item, StructMeta) and not isinstance(item, Field):
        return ClassReference(item)
    if item in [None, ""] or isinstance(item, Field):
        return item
    elif Field in getattr(item, "__mro__", []):
        return item()
    else:
        raise TypeError("Expected a Field/Structure class or Field instance")


class StructureReference(Field):
    """
    A Field that is an embedded structure within other structure. Allows to create hierarchy.
    This is useful if you want to inline your Structure, as opposed to create an explicit
    class for it.
    All the arguments are passed as attributes of the structure. Example:

    .. code-block:: python

        StructureReference(
            _additionalProperties = False,
            id = String,
            name = String
            age = AnyOf[PositiveInt, PositiveFloat]
        )


    Important: Since Typedpy dynamically creates an internal class for it, this
    field cannot be pickled!
    """

    counter = 0

    def __init__(self, **kwargs):
        classname = "StructureReference_" + str(StructureReference.counter)
        StructureReference.counter += 1

        self._newclass = type(classname, (Structure,), kwargs)
        super().__init__(kwargs)

    def __set__(self, instance, value):
        if not isinstance(value, (dict, Structure)):
            raise TypeError(
                f"{self._name}: Expected a dictionary or Structure; got {value}"
            )
        extracted_values = (
            {k: v for (k, v) in value.__dict__.items() if k != "_instantiated"}
            if isinstance(value, (Structure,))
            else value
        )
        newval = self._newclass(**extracted_values)
        super().__set__(instance, newval)

    def __serialize__(self, value):
        raise TypeError(f"{self._name}: StructuredReference Cannot be pickled")

    def __str__(self):
        props = []
        for k, val in sorted(self._newclass.__dict__.items()):
            if val is not None and not k.startswith("_"):
                props.append(f"{k} = {str(val)}")

        propst = f". Properties: {', '.join(props)}" if props else ""
        return f"<Structure{propst}>"


class Number(Field):
    """
    Base class for numerical fields. Based on Json schema draft4.
    Accepts and int or float.

    Arguments:
        multipleOf(int): optional
            The number must be a multiple of this number
        minimum(int or float): optional
            value cannot be lower than this number
        maximum(int or float): optional
            value cannot be higher than this number
        exclusiveMaximum(bool): optional
            marks the maximum threshold above as exclusive

    """

    def __init__(
        self,
        *args,
        multiplesOf=None,
        minimum=None,
        maximum=None,
        exclusiveMaximum=None,
        **kwargs,
    ):
        self.multiplesOf = multiplesOf
        self.minimum = minimum
        self.maximum = maximum
        self.exclusiveMaximum = exclusiveMaximum
        super().__init__(*args, **kwargs)

    def _validate_static(self, value):
        def is_number(val):
            return isinstance(val, (float, int, Decimal))

        def err_prefix():
            return f"{self._name}: Got {wrap_val(value)}; " if self._name else ""

        if not is_number(value):
            raise TypeError(f"{err_prefix()}Expected a number")
        if (
            isinstance(self.multiplesOf, float)
            and int(value / self.multiplesOf) != value / self.multiplesOf
            or isinstance(self.multiplesOf, int)
            and value % self.multiplesOf
        ):
            raise ValueError(
                f"{err_prefix()}Expected a a multiple of {self.multiplesOf}"
            )
        if (is_number(self.minimum)) and self.minimum > value:
            raise ValueError(f"{err_prefix()}Expected a minimum of {self.minimum}")
        if is_number(self.maximum):
            if self.exclusiveMaximum and self.maximum == value:
                raise ValueError(
                    f"{err_prefix()}Expected a maximum of less than {self.maximum}"
                )
            if self.maximum < value:
                raise ValueError(f"{err_prefix()}Expected a maximum of {self.maximum}")

    def _validate(self, value):
        Number._validate_static(self, value)

    def __set__(self, instance, value):
        if not getattr(instance, "_skip_validation", False):
            self._validate(value)
        super().__set__(instance, value)


class Integer(TypedField, Number):
    """
    An extension of :class:`Number` for an integer. Accepts int
    """

    _ty = int

    def _validate(self, value):
        super()._validate(value)
        Number._validate_static(self, value)


class DecimalNumber(Number, SerializableField):
    """
    An extension of :class:`Number` for a Decimal. Accepts anything that can be
    converted to a Decimal.
    It converts the value to a Decimal.
    """

    def __set__(self, instance, value):
        try:
            value = Decimal(value)
        except TypeError as ex:
            raise TypeError(f"{self._name}: {ex.args[0]}") from ex
        except InvalidOperation as ex:
            raise ValueError(f"{self._name}: {ex.args[0]}") from ex

        super().__set__(instance, value)

    def serialize(self, value):
        return float(value)

    @property
    def get_type(self):
        return Decimal


class StructureClass(TypedField):
    _ty = StructMeta


class String(TypedField):
    """
    A string value. Accepts input of `str`

    Arguments:
        minLength(int): optional
            minimal length
        maxLength(int): optional
            maximal lengthr
        pattern(str): optional
            string of a regular expression

    """

    _ty = str

    def __init__(self, *args, minLength=None, maxLength=None, pattern=None, **kwargs):
        self.minLength = minLength
        self.maxLength = maxLength
        self.pattern = pattern
        if self.pattern is not None:
            self._compiled_pattern = re.compile(self.pattern)
        super().__init__(*args, **kwargs)

    def _validate(self, value):
        String._validate_static(self, value)

    def _validate_static(self, value):
        def err_prefix():
            return f"{self._name}: Got {wrap_val(value)}; " if self._name else ""

        if not isinstance(value, str):
            raise TypeError(f"{err_prefix()}Expected a string")
        if self.maxLength is not None and len(value) > self.maxLength:
            raise ValueError(
                f"{err_prefix()}Expected a maximum length of {self.maxLength}"
            )
        if self.minLength is not None and len(value) < self.minLength:
            raise ValueError(
                f"{err_prefix()}Expected a minimum length of {self.minLength}"
            )
        if self.pattern is not None and not self._compiled_pattern.match(value):
            raise ValueError(
                f"{err_prefix()}Does not match regular expression: {wrap_val(self.pattern)}"
            )

    def __set__(self, instance, value):
        self._validate(value)
        super().__set__(instance, value)


class Function(Field):
    """
    A function or method. Note that this can't be any callable (it can't be a class,
     for example), but a real function
    """

    _bound_method_type = type(Field().__init__)

    def __set__(self, instance, value):
        def is_function(f):
            return type(f) in {
                type(lambda x: x),
                type(open),
                Function._bound_method_type,
            }

        def err_prefix():
            return f"{self._name}: Got {wrap_val(value)}; " if self._name else ""

        if not is_function(value):
            raise TypeError(f"{err_prefix()}Expected a function")
        super().__set__(instance, value)


class Generator(TypedField):
    """
    A Python generator. Not serializable.
    """

    _ty = typing.Generator


class Anything(Field):
    """
    A field that can contain anything (similar to "any" in Typescript).
    Example:

    .. code-block:: python

        class Foo(Structure):
            i = Integer
            some_content = Anything

        # now we can assign anything to some_content property:
        Foo(i=5, some_content = "whatever")
        Foo(i=5, some_content = [1,2,3])
        Foo(i=5, some_content = Bar())

    """

    pass


class Float(TypedField, Number):
    """
    An extension of :class:`Number` for a float
    """

    _ty = float

    def _validate(self, value):
        super()._validate(value)
        Number._validate_static(self, value)


class Boolean(TypedField):
    """
    Value of type bool. True or False.
    """

    _ty = bool

    def __set__(self, instance, value):
        mapping = {"True": True, "False": False}
        value = mapping[value] if value in mapping else value
        super().__set__(instance, value)

    def _validate(self, value):
        def err_prefix():
            return f"{self._name}: " if self._name else ""

        if value not in {"True", "False", True, False}:
            raise TypeError(f"{err_prefix()}Expected {self._ty}; Got {wrap_val(value)}")


class Positive(Number):
    """
    An extension of :class:`Number`. Requires the number to be positive
    """

    def __set__(self, instance, value):
        if value <= 0:
            raise ValueError(f"{self._name}: Got {value}; Expected a positive number")
        super().__set__(instance, value)


class NonPositive(Number):
    """
    An extension of :class:`Number`. Requires the number to be negative or 0
    """

    def __set__(self, instance, value):
        if value > 0:
            raise ValueError(
                f"{self._name}: Got {value}; Expected a negative number or 0"
            )
        super().__set__(instance, value)


class Negative(Number):
    """
    An extension of :class:`Number`. Requires the number to be negative
    """

    def __set__(self, instance, value):
        if value >= 0:
            raise ValueError(f"{self._name}: Got {value}; Expected a negative number")
        super().__set__(instance, value)


class NonNegative(Number):
    """
    An extension of :class:`Number`. Requires the number to be positive or 0
    """

    def __set__(self, instance, value):
        if value < 0:
            raise ValueError(
                f"{self._name}: Got {value}; Expected a positive number or 0"
            )
        super().__set__(instance, value)


class PositiveFloat(Float, Positive):
    """
    An combination of :class:`Float` and :class:`Positive`
    """

    pass


class PositiveInt(Integer, Positive):
    """
    An combination of :class:`Integer` and :class:`Positive`
    """

    pass


class NegativeFloat(Float, Negative):
    """
    An combination of :class:`Float` and :class:`Negative`
    """

    pass


class NegativeInt(Integer, Negative):
    """
    An combination of :class:`Integer` and :class:`Negative`
    """

    pass


class NonPositiveFloat(Float, NonPositive):
    """
    An combination of :class:`Float` and :class:`NonPositive`
    """

    pass


class NonPositiveInt(Integer, NonPositive):
    """
    An combination of :class:`Integer` and :class:`NonPositive`
    """

    pass


class NonNegativeFloat(Float, NonNegative):
    """
    An combination of :class:`Float` and :class:`NonNegative`
    """

    pass


class NonNegativeInt(Integer, NonNegative):
    """
    An combination of :class:`Integer` and :class:`NonNegative`
    """

    pass


class _IteratorProxyMixin:
    class ListIteratorProxy:
        def __init__(self, the_list):
            self.the_list = the_list
            self.index = 0

        def __next__(self):
            if len(self.the_list) > self.index:
                self.index += 1
                return self.the_list[self.index - 1]
            raise StopIteration


class _ListStruct(list, ImmutableMixin, _IteratorProxyMixin):
    """
    This is a useful wrapper for the content of list in an Array field.
    It ensures that an update of the form:
     mystruct.my_array[i] = new_val
    Will not bypass the validation of the Array.
    """

    def __init__(self, array: Field, struct_instance: Structure, mylist, name: str):
        self._field_definition = array
        self._instance = struct_instance
        self._name = name
        super().__init__(self._get_defensive_copy_if_needed(mylist))

    def __setitem__(self, key, value):
        self._raise_if_immutable()
        copied = self[:]
        copied.__setitem__(key, value)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)

    def __getitem__(self, item):
        val = super().__getitem__(item)
        return self._get_defensive_copy_if_needed(val)

    def __iter__(self):
        if self._is_immutable():
            return _IteratorProxyMixin.ListIteratorProxy(self)
        return super(_ListStruct, self).__iter__()

    def append(self, value):
        self._raise_if_immutable()
        copied = self[:]
        copied.append(value)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
        super().append(value)

    def extend(self, value):
        self._raise_if_immutable()
        copied = self[:]
        copied.extend(value)
        if getattr(self, "_instance", None):
            setattr(
                self._instance, getattr(self._field_definition, "_name", None), copied
            )

    def insert(self, index: int, value):
        self._raise_if_immutable()
        copied = self[:]
        copied.insert(index, value)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)

    def remove(self, ind):
        self._raise_if_immutable()
        copied = self[:]
        copied.remove(ind)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)

    def copy(self):
        copied = super(_ListStruct, self).copy()
        return deepcopy(copied) if self._is_immutable() else copied

    def clear(self) -> None:
        self._raise_if_immutable()
        setattr(self._instance, getattr(self._field_definition, "_name", None), [])

    def pop(self, index: int = -1):
        self._raise_if_immutable()
        copied = self[:]
        res = copied.pop(index)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
        return res

    def __getstate__(self):
        return {
            "the_instance": self._instance,
            "the_array": self._field_definition,
            "the_name": self._name,
            "the_values": self[:],
        }

    def __deepcopy__(self, memo={}):
        vals = [deepcopy(v) for v in self[:]]
        instance_id = id(self._instance)
        return _ListStruct(
            array=deepcopy(self._field_definition),
            struct_instance=memo.get(instance_id, self._instance),
            mylist=vals,
            name=self._name,
        )

    def __setstate__(self, state):
        self._name = state["the_name"]
        self._field_definition = state["the_array"]
        self._instance = state["the_instance"]
        super().__init__(state["the_values"])


class _DequeStruct(deque, ImmutableMixin, _IteratorProxyMixin):
    """
    This is a useful wrapper for the content of list in an Deque field.
    It ensures that an update of the form:
     mystruct.my_array[i] = new_val
    Will not bypass the validation of the Array.
    """

    def __init__(
        self,
        deq: Field = None,
        struct_instance: Structure = None,
        mydeque=None,
        name: str = None,
    ):
        self._field_definition = deq
        self._instance = struct_instance
        self._name = name
        if mydeque is not None:
            super().__init__(self._get_defensive_copy_if_needed(mydeque))

    def __setitem__(self, key, value):
        self._raise_if_immutable()
        copied = deque(self)
        copied[key] = value
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)

    def __getitem__(self, item):
        val = super().__getitem__(item)
        return self._get_defensive_copy_if_needed(val)

    def __iter__(self):
        if self._is_immutable():
            return _ListStruct.ListIteratorProxy(self)
        return super(_DequeStruct, self).__iter__()

    def append(self, x):
        self._raise_if_immutable()
        copied = deque(self)
        copied.append(x)
        if self._field_definition:  # Python 3.6
            setattr(
                self._instance, getattr(self._field_definition, "_name", None), copied
            )
        super().append(x)

    def appendleft(self, x):
        self._raise_if_immutable()
        copied = deque(self)
        copied.appendleft(x)
        if self._field_definition:  # Python 3.6
            setattr(
                self._instance, getattr(self._field_definition, "_name", None), copied
            )
        super().append(x)

    def extend(self, iterable: Iterable):
        self._raise_if_immutable()
        copied = deque(self)
        copied.extend(iterable)
        if getattr(self, "_instance", None):
            setattr(
                self._instance, getattr(self._field_definition, "_name", None), copied
            )

    def extendleft(self, iterable: Iterable):
        self._raise_if_immutable()
        copied = deque(self)
        copied.extendleft(iterable)
        if getattr(self, "_instance", None):
            setattr(
                self._instance, getattr(self._field_definition, "_name", None), copied
            )

    def insert(self, x: int, i):
        self._raise_if_immutable()
        copied = deque(self)
        copied.insert(x, i)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)

    def remove(self, value):
        self._raise_if_immutable()
        copied = deque(self)
        copied.remove(value)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)

    def copy(self):
        copied = deque(self)
        return deepcopy(copied) if self._is_immutable() else copied

    def clear(self) -> None:
        self._raise_if_immutable()
        setattr(self._instance, getattr(self._field_definition, "_name", None), deque())

    def pop(self, *args, **kwargs):
        self._raise_if_immutable()
        copied = deque(self)
        res = copied.pop()
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
        return res

    def popleft(self):
        self._raise_if_immutable()
        copied = deque(self)
        res = copied.popleft()
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
        return res

    def rotate(self, n: int) -> None:  # pylint: disable=signature-differs
        self._raise_if_immutable()
        # no need to validate again
        super().rotate(n)

    def reverse(self) -> None:
        self._raise_if_immutable()
        # no need to validate again
        super().reverse()

    def __getstate__(self):
        return {
            "the_instance": self._instance,
            "field_def": self._field_definition,
            "the_name": self._name,
            "the_values": deque(self),
        }

    def __deepcopy__(self, memo={}):
        vals = [deepcopy(v) for v in self.copy()]
        instance_id = id(self._instance)
        return _DequeStruct(
            deq=deepcopy(self._field_definition),
            struct_instance=memo.get(instance_id, self._instance),
            mydeque=vals,
            name=self._name,
        )

    def __reduce__(self):
        res = super().__reduce__()
        return res[0], res[1], self.__getstate__(), res[3]

    def __setstate__(self, state):
        self._name = state["the_name"]
        self._field_definition = state["field_def"]
        self._instance = state["the_instance"]
        super().__init__(state["the_values"])


class _DictStruct(dict, ImmutableMixin):
    """
    This is a useful wrapper for the content of dict in an Map field.
    It ensures that an update of the form:
     mystruct.my_map[i] = new_val, or
     mystruct.my_map.update(some_dict)

    ...will not bypass the validation of the Map.
    """

    def __init__(self, the_map, struct_instance, mydict, name):
        self._field_definition = the_map
        self._instance = struct_instance
        self._name = name
        super().__init__(mydict)

    def __setitem__(self, key, value):
        super()._raise_if_immutable()
        copied = self.copy()
        copied.__setitem__(key, value)
        if getattr(self, "_instance", None):
            setattr(
                self._instance, getattr(self._field_definition, "_name", None), copied
            )

        super().__setitem__(key, value)

    def __getitem__(self, item):
        val = super().__getitem__(item)
        return self._get_defensive_copy_if_needed(val)

    def copy(self):
        copied = super(_DictStruct, self).copy()
        return deepcopy(copied) if self._is_immutable() else copied

    def __deepcopy__(self, memo={}):
        new_dict = {deepcopy(k): deepcopy(v) for k, v in self.items()}
        instance_id = id(self._instance)
        return _DictStruct(
            the_map=self._field_definition,
            struct_instance=memo.get(instance_id, self._instance),
            mydict=new_dict,
            name=self._name,
        )

    def items(self):
        return ((k, self._get_defensive_copy_if_needed(v)) for k, v in super().items())

    def values(self):
        return (self._get_defensive_copy_if_needed(v) for v in super().values())

    def __delitem__(self, key):
        self._raise_if_immutable()
        copied = self.copy()
        del copied[key]
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)

    def update(self, *args, **kwargs):
        self._raise_if_immutable()
        copied = self.copy()
        res = copied.update(*args, **kwargs)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
        return res

    def pop(self, k):
        self._raise_if_immutable()
        copied = self.copy()
        res = copied.pop(k)
        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
        return res

    def clear(self) -> None:
        self._raise_if_immutable()
        setattr(self._instance, getattr(self._field_definition, "_name", None), {})

    def __getstate__(self):
        return {
            "_instance": self._instance,
            "_map": self._field_definition,
            "mydict": self.copy(),
            "_name": self._name,
        }

    def __setstate__(self, state):
        self._field_definition = state["_map"]
        self._instance = state["_instance"]
        self._name = state["_name"]
        super().__init__(state["mydict"])


class _CollectionMeta(_FieldMeta):
    def __getitem__(cls, item):
        def validate_and_get_field(val):
            return _FieldMeta.__getitem__(cls, val)

        if isinstance(item, tuple):
            items = [validate_and_get_field(it) for it in item]
            return cls(items=items)  # pylint: disable=E1120, E1123
        return cls(items=validate_and_get_field(item))  # pylint: disable=E1120, E1123


class _JSONSchemaDraft4ReuseMeta(_FieldMeta):
    def __getitem__(cls, item):
        def validate_and_get_field(val):
            return _FieldMeta.__getitem__(cls, val)

        if isinstance(item, tuple):
            fields = [validate_and_get_field(it) for it in item]
            return cls(fields)  # pylint: disable=E1120, E1123
        return cls([validate_and_get_field(item)])  # pylint: disable=E1120, E1123


class ContainNestedFieldMixin(Field):
    def _set_immutable(self, immutable: bool):
        items = getattr(self, "items", None)
        super()._set_immutable(immutable)
        if isinstance(items, Field):
            items._set_immutable(immutable)
        elif isinstance(items, (list, tuple)):
            for item in items:
                if isinstance(item, Field):
                    item._set_immutable(immutable)


class SizedCollection:
    def __init__(self, *args, minItems=None, maxItems=None, **kwargs):
        self.minItems = minItems
        self.maxItems = maxItems
        super().__init__(*args, **kwargs)

    def validate_size(self, items, name):
        if self.minItems is not None and len(items) < self.minItems:
            raise ValueError(
                f"{name}: Expected length of at least {self.minItems}; Got {items}"
            )
        if self.maxItems is not None and len(items) > self.maxItems:
            raise ValueError(
                f"{name}: Expected length of at most {self.maxItems}; Got {items}"
            )

    # This is needed to hack the type check of typing.Optional, to allow the following syntax:
    #   x: Optional[Array[Integer]]
    def __call__(self, *args, **kwargs):
        return self


class Set(
    SizedCollection, ContainNestedFieldMixin, TypedField, metaclass=_CollectionMeta
):
    """
    A set collection. Accepts input of type `set`

    Arguments:
        minItems(int): optional
            minimal size
        maxItems(int): optional
            maximal size
        items(:class:`Field` or :class:`Structure`): optional
            The type of the content, can be a predefined :class:`Structure`,
            :class:`Field` or an arbitrary class. In case of an arbitrary
            class, an implicit Field class will be created for it behind the
            scenes. Always prefer an Explicit Typedpy  :class:`Structure` or
            :class:`Field`  if you can.

    Examples:

    .. code-block:: python

        Set[String]
        Set(items=Integer(maximum=10), maxItems = 10)

        # let's assume we defined a Structure 'Person', then we can use it too:
        Set[Person]


    """

    _ty = set

    def __init__(self, *args, items=None, **kwargs):
        self.items = _map_to_field(items)

        if isinstance(self.items, TypedField) and not getattr(
            getattr(self.items, "_ty"), "__hash__"
        ):
            raise TypeError(
                f"Set element of type {getattr(self.items, '_ty')} is not hashable"
            )
        super().__init__(*args, **kwargs)
        self._set_immutable(getattr(self, "_immutable", False))

    @property
    def get_type(self):
        if (
            not isinstance(self.items, (list, tuple))
            and self.items
            and python_ver_atleast_39
        ):
            return set[self.items.get_type]
        return set

    def __set__(self, instance, value):
        cls = self.__class__._ty
        if not isinstance(value, cls):
            raise TypeError(f"{self._name}: Got {wrap_val(value)}; Expected {cls}")
        self.validate_size(value, self._name)
        if self.items is not None:
            setattr(self.items, "_name", self._name)
            res = []
            for val in value:
                temp_st = Structure()
                self.items.__set__(temp_st, val)
                res.append(getattr(temp_st, getattr(self.items, "_name")))
            value = cls(res)
        super().__set__(instance, value)


class Map(
    SizedCollection, ContainNestedFieldMixin, TypedField, metaclass=_CollectionMeta
):
    """
    A map/dictionary collection. Accepts input of type `dict`

    Arguments:
        minItems(int): optional
            minimal size
        maxItems(int): optional
            maximal size
        items(tuple of 2 :class:`Field` or :class:`Structure`  elements): optional
            The first element is the Field for keys, the second is for values.
            Examples:

            .. code-block:: python

                age_by_name = Map[String, PositiveInt]
                # Let's assume we defined a Structure "Person"
                person_by_id = Map[String, Person]
                # even Structure reference is supported for keys!
                id_by_person = Map[Person, String]
                id_by_person = Map[Person, String]

    """

    _ty = dict

    def __init__(self, *args, items=None, **kwargs):
        if items is not None and (
            not isinstance(items, (tuple, list)) or len(items) != 2
        ):
            raise TypeError("items is expected to be a list/tuple of two fields")
        if items is None:
            self.items = None
        else:
            self.items = []
            for item in items:
                self.items.append(_map_to_field(item))
            key_field = self.items[0]
            if isinstance(key_field, TypedField) and not getattr(
                getattr(key_field, "_ty"), "__hash__"
            ):
                raise TypeError(
                    f"Key field of type {key_field}, with underlying type of {getattr(key_field, '_ty')} "
                    "is not hashable"
                )
        self._custom_deep_copy_implementation = True
        super().__init__(*args, **kwargs)
        self._set_immutable(getattr(self, "_immutable", False))

    def __set__(self, instance, value):
        if not isinstance(value, dict):
            raise TypeError(f"{self._name}: Expected a dict")
        self.validate_size(value, self._name)

        if self.items is not None:
            key_field, value_field = self.items[0], self.items[1]
            setattr(key_field, "_name", self._name + "_key")
            setattr(value_field, "_name", self._name + "_value")
            res = OrderedDict()
            for key, val in value.items():
                temp_st = Structure()
                key_field.__set__(temp_st, key)
                value_field.__set__(temp_st, val)

                res[getattr(temp_st, getattr(key_field, "_name"))] = getattr(
                    temp_st, getattr(value_field, "_name")
                )
                value = res

        super().__set__(instance, _DictStruct(self, instance, value, self._name))


class Array(
    SizedCollection, ContainNestedFieldMixin, TypedField, metaclass=_CollectionMeta
):
    """
    An Array field, similar to a list. Supports the properties in JSON schema draft 4.
    Expected input is of type `list`.

    Arguments:
        minItems(int): optional
            minimal size
        maxItems(int): optional
            maximal size
        unqieItems(bool): optional
            are elements required to be unique?
        additionalItems(bool): optional
            Relevant in case items parameter is a list of Fields. Is it allowed to have additional
            elements beyond the ones defined in "items"?
        items(a :class:`Field` or :class:`Structure`, or a list/tuple of :class:`Field` or :class:`Structure`): optional
            Describes the fields of the elements.
            If a items if a :class:`Field`, then it applies to all items.
            If a items is a list, then every element in the content is expected to be
            of the corresponding field in items.
            Examples:

            .. code-block:: python

                names = Array[String]
                names = Array[String(minLengh=3)]
                names = Array(minItems=5, items=String)
                my_record = Array(items=[String, Integer(minimum=5), String])
                my_lists = Array[Array[Integer]]
                my_structs = Array[StructureReference(a=Integer, b=Float)]
                # Let's say we defined a Structure "Person"
                people = Array[Person]

                # Assume Foo is an arbitrary (non-Typedpy) class
                foos = Array[Foo]

    """

    _ty = list

    def __init__(
        self, *args, items=None, uniqueItems=None, additionalItems=None, **kwargs
    ):
        """
        Constructor
        :param args: pass-through
        :param items: either a single field, which will be enforced for all elements, or a list
         of fields which enforce the elements with the correspondent index
        :param uniqueItems: are elements required to be unique?
        :param additionalItems: Relevant if "items" is a list. Is it allowed to have additional
        elements beyond the ones defined in "items"?
        :param kwargs: pass-through
        """
        self.uniqueItems = uniqueItems
        self.additionalItems = additionalItems
        if isinstance(items, list):
            self.items = []
            for item in items:
                self.items.append(_map_to_field(item))
        else:
            self.items = _map_to_field(items)
        super().__init__(*args, **kwargs)
        self._set_immutable(getattr(self, "_immutable", False))

    @property
    def get_type(self):
        if (
            not isinstance(self.items, (list, tuple))
            and self.items
            and python_ver_atleast_39
        ):
            return list[self.items.get_type]
        return list

    def __set__(self, instance, value):
        verify_type_and_uniqueness(list, value, self._name, self.uniqueItems)
        self.validate_size(value, self._name)
        if self.items is not None:
            if isinstance(self.items, Field):
                setattr(self.items, "_name", self._name)
                res = []
                for i, val in enumerate(value):
                    temp_st = Structure()
                    setattr(self.items, "_name", self._name + f"_{str(i)}")
                    self.items.__set__(temp_st, val)
                    res.append(getattr(temp_st, getattr(self.items, "_name")))
                value = res
            elif isinstance(self.items, list):
                additional_properties_forbidden = self.additionalItems is False

                if not getattr(instance, "_skip_validation", False):
                    if len(self.items) > len(value) or (
                        additional_properties_forbidden and len(self.items) > len(value)
                    ):
                        raise ValueError(
                            f"{self._name}: Got {value}; Expected an array of length {len(self.items)}"
                        )
                temp_st = Structure()
                temp_st._skip_validation = getattr(instance, "_skip_validation", False)
                res = []
                for ind, item in enumerate(self.items):
                    if ind >= len(value):
                        continue
                    setattr(item, "_name", self._name + f"_{str(ind)}")
                    item.__set__(temp_st, value[ind])
                    res.append(getattr(temp_st, getattr(item, "_name")))
                res += value[len(self.items) :]
                value = res

        super().__set__(instance, _ListStruct(self, instance, value, self._name))


class Deque(
    SizedCollection, ContainNestedFieldMixin, TypedField, metaclass=_CollectionMeta
):
    """
    An collections.deque field. Supports the properties in JSON schema draft 4.
    Expected input is of type `collections.deque`.

    Arguments:
        minItems(int): optional
            minimal size
        maxItems(int): optional
            maximal size
        unqieItems(bool): optional
            are elements required to be unique?
        additionalItems(bool): optional
            Relevant in case items parameter is a list of Fields. Is it allowed to have additional
            elements beyond the ones defined in "items"?
        items(a :class:`Field` or :class:`Structure`, or a list/tuple of :class:`Field` or :class:`Structure`): optional
            Describes the fields of the elements.
            If a items if a :class:`Field`, then it applies to all items.
            If a items is a list, then every element in the content is expected to be
            of the corresponding field in items.
            Examples:

            .. code-block:: python

                names = Deque[String]
                names = Deque[String(minLengh=3)]
                names = Deque(minItems=5, items=String)
                my_record = Deque(items=[String, Integer(minimum=5), String])
                my_lists = Deque[Array[Integer]]
                my_structs = Deque[StructureReference(a=Integer, b=Float)]
                # Let's say we defined a Structure "Person"
                people = Deque[Person]

                # Assume Foo is an arbitrary (non-Typedpy) class
                foos = Deque[Foo]

    """

    _ty = deque

    def __init__(
        self, *args, items=None, uniqueItems=None, additionalItems=None, **kwargs
    ):
        """
        Constructor
        :param args: pass-through
        :param items: either a single field, which will be enforced for all elements, or a list
         of fields which enforce the elements with the correspondent index
        :param uniqueItems: are elements required to be unique?
        :param additionalItems: Relevant if "items" is a list. Is it allowed to have additional
        elements beyond the ones defined in "items"?
        :param kwargs: pass-through
        """
        self.uniqueItems = uniqueItems
        self.additionalItems = additionalItems
        if isinstance(items, list):
            self.items = []
            for item in items:
                self.items.append(_map_to_field(item))
        else:
            self.items = _map_to_field(items)
        super().__init__(*args, **kwargs)
        self._set_immutable(getattr(self, "_immutable", False))

    def __set__(self, instance, value):
        verify_type_and_uniqueness(deque, value, self._name, self.uniqueItems)
        self.validate_size(value, self._name)
        if self.items is not None:
            if isinstance(self.items, Field):
                setattr(self.items, "_name", self._name)
                res = deque()
                for i, val in enumerate(value):
                    temp_st = Structure()
                    setattr(self.items, "_name", self._name + f"_{str(i)}")
                    self.items.__set__(temp_st, val)
                    res.append(getattr(temp_st, getattr(self.items, "_name")))
                value = res
            elif isinstance(self.items, list):
                additional_properties_forbidden = self.additionalItems is False

                if not getattr(instance, "_skip_validation", False):
                    if len(self.items) > len(value) or (
                        additional_properties_forbidden and len(self.items) > len(value)
                    ):
                        raise ValueError(
                            f"{self._name}: Got {value}; Expected an deque of length {len(self.items)}"
                        )
                temp_st = Structure()
                temp_st._skip_validation = getattr(instance, "_skip_validation", False)
                res = deque()
                for ind, item in enumerate(self.items):
                    if ind >= len(value):
                        continue
                    setattr(item, "_name", self._name + f"_{str(ind)}")
                    item.__set__(temp_st, value[ind])
                    res.append(getattr(temp_st, getattr(item, "_name")))
                for i in range(len(self.items), len(value)):
                    res.append(value[i])
                value = res

        super().__set__(instance, _DequeStruct(self, instance, value, self._name))


def verify_type_and_uniqueness(the_type, value, name, has_unique_items):
    if not isinstance(value, the_type):
        raise TypeError(f"{name}: Got {wrap_val(value)}; Expected {str(the_type)}")
    if has_unique_items:
        unique = reduce(
            lambda unique_vals, x: unique_vals.append(x) or unique_vals
            if x not in unique_vals
            else unique_vals,
            value,
            [],
        )
        if len(unique) < len(value):
            raise ValueError(f"{name}: Got {wrap_val(value)}; Expected unique items")


class Tuple(ContainNestedFieldMixin, TypedField, metaclass=_CollectionMeta):
    """
    A tuple field, supports unique items option.
       Expected input is of type `tuple`.

    Arguments:

        unqieItems(`bool`): optional
            are elements required to be unique?

        items(`list`/`tuple` of :class:`Field` or :class:`Structure`): optional
            Describes the fields of the elements.
            Every element in the content is expected to be
            of the corresponding :class:`Field` in items.


    Examples:

    .. code-block:: python

        # a is a tuple of exactly 2 strings that are different from each other.
        a = Tuple(uniqueItems=True, items = [String, String])

        # b is a tuple of 3: string, string and a number up to 10.
        b = Tuple(items = [String, String, Number(maximum=10)])

        # c is a tuple of 3: integer, string, float.
        c = Tuple[Integer, String, Float]

        # The following define a tuple of any number of Integers
        d = Tuple[Integer]

        # It can also contain other structures:
        # Assume we have something like: class Foo(Structure): pass
        # e is a tuple of any number of Integers or Foo instances
        e = Tuple[AnyOf[Integer, Foo]]

        # It can also have arbitrary class
        class MyCustomClass: pass
        Tuple[MyCustomClass]
    """

    _ty = tuple

    def __init__(self, *args, items, uniqueItems=None, **kwargs):
        """
        Constructor
        :param args: pass-through
        :param items: either a single field, which will be enforced for all elements, or a list
         of fields which enforce the elements with the correspondent index
        :param uniqueItems: are elements required to be unique?
        :param kwargs: pass-through
        """
        self.uniqueItems = uniqueItems
        if isinstance(items, (list, tuple)):
            self.items = []
            for item in items:
                if isinstance(item, Field):
                    self.items.append(item)
                elif Field in item.__mro__:
                    self.items.append(item())
                else:
                    raise TypeError("Expected a Field class or instance")
        elif isinstance(items, (Field,)) or Field in items.__mro__:
            self.items = [items]
        else:
            raise TypeError("Expected a list/tuple of Fields or a single Field")
        super().__init__(*args, **kwargs)
        self._set_immutable(getattr(self, "_immutable", False))

    @property
    def get_type(self):
        if self.items and python_ver_atleast_39:
            if not isinstance(self.items, (list, tuple)):
                return tuple[self.items.get_type]
            if len(self.items) == 2:
                return tuple[self.items[0].get_type, self.items[1].get_type]
            if len(self.items) == 3:
                return tuple[
                    self.items[0].get_type,
                    self.items[1].get_type,
                    self.items[2].get_type,
                ]
        return tuple

    def __set__(self, instance, value):
        verify_type_and_uniqueness(tuple, value, self._name, self.uniqueItems)
        if len(self.items) != len(value) and len(self.items) > 1:
            raise ValueError(
                f"{self._name}: Got {wrap_val(value)}; Expected a tuple of length {len(self.items)}"
            )

        temp_st = Structure()
        res = []
        items = self.items if len(self.items) > 1 else self.items * len(value)
        for ind, item in enumerate(items):
            setattr(item, "_name", self._name + f"_{str(ind)}")
            item.__set__(temp_st, value[ind])
            res.append(getattr(temp_st, getattr(item, "_name")))
            res += value[len(items) :]
        value = tuple(res)

        super().__set__(instance, value)


class Sized(Field):
    """
    The length of the value is limited to be at most the maximum given.
    The value can be any iterable.

        Arguments:

            maxlen(`int`):
                maximum length

    """

    def __init__(self, *args, maxlen, **kwargs):
        self.maxlen = maxlen
        super().__init__(*args, **kwargs)

    def __set__(self, instance, value):
        if len(value) > self.maxlen:
            raise ValueError(
                f"{self._name}: Got {wrap_val(value)}; Expected a length up to {self.maxlen}"
            )
        super().__set__(instance, value)


class SizedString(String, Sized):
    pass


def _str_for_multioption_field(instance):
    name = instance.__class__.__name__
    if instance.get_fields():
        fields_st = ", ".join([str(field) for field in instance.get_fields()])
        propst = f" [{fields_st}]"
    else:
        propst = ""
    return f"<{name}{propst}>"


class MultiFieldWrapper:
    """
    An abstract base class for AllOf, AnyOf, OneOf, etc.
    It provides flexibility in reading the "fields" argument.
    """

    def __init__(self, *arg, fields, **kwargs):
        if isinstance(fields, list):
            self._fields = []
            for item in fields:
                self._fields.append(_map_to_field(item))
        else:
            raise TypeError("Expected a Field class or instance")
        super().__init__(*arg, **kwargs)

    def get_fields(self):
        return self._fields


class AllOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
    """
    Content must adhere to all requirements in the fields arguments.
    Arguments:

        fields( `list` of :class:`Field`): optional
        the content should match all of the fields in the list

    Example:

    .. code-block:: python

        AllOf[Number(maximum=20, minimum=-10), Integer, Positive]

    """

    def __init__(self, fields):
        super().__init__(fields=fields)

    def __set__(self, instance, value):
        for field in self.get_fields():
            setattr(field, "_name", self._name)
            field.__set__(instance, value)
        super().__set__(instance, value)

    def __str__(self):
        return _str_for_multioption_field(self)


class AnyOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
    """
    Content must adhere to one or more of the requirements in the fields arguments.
    Arguments:

        fields( `list` of :class:`Field`): optional
        the content should match at least one of the fields in the list

    Example:

    .. code-block:: python

       AnyOf[Number(maximum=20, minimum=-10), Integer, Positive, String]

    """

    def __init__(self, fields):
        super().__init__(fields=fields)
        if fields:
            for f in fields:
                if isinstance(f, NoneField):
                    self._is_optional = True
        else:
            raise TypeError("AnyOf definition must include at least one field option")

    def __set__(self, instance, value):
        matched = False
        for field in self.get_fields():
            setattr(field, "_name", self._name)
            try:
                field.__set__(instance, value)
                matched = True
            except TypeError:
                pass
            except ValueError:
                pass
        if not matched:
            prefix = f"{self._name}: " if self._name else ""
            raise ValueError(
                f"{prefix}{wrap_val(value)} Did not match any field option"
            )
        super().__set__(instance, getattr(instance, self._name))

    def __str__(self):
        return _str_for_multioption_field(self)


class OneOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
    """
    Content must adhere to one, and only one, of the requirements in the fields arguments.
    Arguments:

        fields( `list` of :class:`Field`): optional
        the content should match one, and only one, of the fields in the list

    Example:

    .. code-block:: python

        OneOf[Number(maximum=20, minimum=-10), Integer, Positive, String]

    """

    def __init__(self, fields):
        super().__init__(fields=fields)

    def __set__(self, instance, value):
        matched = 0
        for field in self.get_fields():
            setattr(field, "_name", self._name)
            try:
                field.__set__(instance, value)
                matched += 1
            except TypeError:
                pass
            except ValueError:
                pass
        if not matched:
            raise ValueError(
                f"{self._name}: Got {value}; Did not match any field option"
            )
        if matched > 1:
            raise ValueError(
                f"{self._name}: Got {value}; Matched more than one field option"
            )
        super().__set__(instance, value)

    def __str__(self):
        return _str_for_multioption_field(self)


class NotField(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
    """
    Content *must not* adhere to any of the requirements in the fields arguments.
    Arguments:

        fields( `list` of :class:`Field`): optional
            the content must not match any of the fields in the lists

    Examples:

    .. code-block:: python

        NotField([Number(multiplesOf=5, maximum=20, minimum=-10), String])
        NotField[Positive]

    """

    def __init__(self, fields):
        super().__init__(fields=fields)

    def __set__(self, instance, value):
        for field in self.get_fields():
            setattr(field, "_name", self._name)
            try:
                field.__set__(instance, value)
            except TypeError:
                pass
            except ValueError:
                pass
            else:
                raise ValueError(
                    f"{self._name}: Got {wrap_val(value)}; Expected not to match any field definition"
                )
        super().__set__(instance, value)

    def __str__(self):
        return _str_for_multioption_field(self)


class ImmutableSet(Set, ImmutableField):
    """
    An immutable  :class:`Set`. Internally implemented by a Python frozenset, so it does not have
    any mutation methods. This makes it more developer-friendly.
    """

    _ty = frozenset

    def __set__(self, instance, value):
        if not isinstance(value, (set, frozenset)):
            raise TypeError(f"{self._name}: Got {wrap_val(value)}; Expected {set}")
        self.validate_size(value, self._name)
        if self.items is not None:
            temp_st = Structure()
            setattr(self.items, "_name", self._name)
            res = set()
            for val in value:
                if getattr(self, "_immutable", False):
                    temp_st = Structure()
                self.items.__set__(temp_st, val)
                res.add(getattr(temp_st, getattr(self.items, "_name")))
                value = res
        corrected_value = value if isinstance(value, frozenset) else frozenset(value)
        super().__set__(instance, corrected_value)


class ImmutableMap(ImmutableField, Map):
    """
    An immutable version of :class:`Map`
    """

    pass


class ImmutableArray(ImmutableField, Array):
    """
    An immutable version of :class:`Array`
    """

    pass


class ImmutableDeque(ImmutableField, Deque):
    """
    An immutable version of :class:`Deque`
    """

    pass


class ImmutableString(ImmutableField, String):
    """
    An immutable version of :class:`String`
    """

    pass


class ImmutableNumber(ImmutableField, Number):
    """
    An immutable version of :class:`Number`
    """

    pass


class ImmutableInteger(ImmutableField, Integer):
    """
    An immutable version of :class:`Integer`
    """

    pass


class ImmutableFloat(ImmutableField, Float):  # pylint: disable=
    """
    An immutable version of :class:`Float`
    """

    pass


class ExceptionField(TypedField, SerializableField):
    """
    As Exception. This is serialized as the string representation of the exception.
    It does not support deserialization.
    """

    _ty = Exception

    def serialize(self, value):
        return f"{value.__class__.__name__}: {str(value)}"


class FunctionCall(Structure):
    """
    Structure that represents a function call for the purpose of serialization mappers: \
    Includes the function to be called, and a list of keys of positional string arguments.
    This is not a generic function call.

    Arguments:
        func(Function):
            the function to be called. Note this must be a real function, not any callable.
        args(Array[String]): optional
            the keys of the arguments to be used as positional arguments for the function call
    """

    func = Function
    args = Array[String]
    _required = ["func"]



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    long_description = readme.read()

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
]

setup(
    name="typedpy",
    packages=find_packages(include=["typedpy", "typedpy.scripts"]),
    entry_points={
        'console_scripts': [
            'create-stub=typedpy.scripts.create_stub:main',
            'create-stubs-for-dir=typedpy.scripts.create_stubs:main',
        ],
    },
    install_requires=[],
    author="Danny Loya",
    author_email="dan.loya@gmail.com",
    classifiers=classifiers,
    description="Type-safe Python",
    long_description_content_type="text/markdown",
    license="MIT",
    long_description=long_description,
    url="http://github.com/loyada/typedpy",
    download_url ="https://github.com/loyada/typedpy/archive/v2.15.39.tar.gz",
    keywords=['testing', 'type-safe', 'strict', 'schema', 'validation'],
    version='2.15.39'
)

# coverage run --source=typedpy/ setup.py test
# coverage html
# load in browser from coverage_html_report

# pylint --rcfile=setup.cfg typedpy
#    or
# python3 setup.py lint
